/*
 * Dapple ][ Emulator  Version 0.24
 * Copyright (C) 2002, 2003 Dapple ][ Development.  All rights reserved.
 *
 * Component:  VIDEO: routines graphic output
 * Revision:   (0.24) 2003.0226
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <string.h>
#include "asmlib.h"
#include "dapple.h"
#include "font.h"

/* dapple.c */
extern unsigned int window;	/* handle to window */
extern unsigned char messageflag;
extern unsigned char appletype;
extern Charset inicharmode;
void setmessage(unsigned char *message);

/* memory.c */
extern unsigned char memram[0x28000];
extern unsigned char memstore80;


/* Methods:

   - virtsetmode
   - virtcacheinit
   - virtsetcharset
   - virtgetcharset
   - virtsettextmode
   - virtgettextmode
   - virtsethresmode
   - virtgethresmode
   - virtsetdhrsmode
   - virtgetdhrsmode
   - virtsetmonochrome
   - virtgetmonochrome
   - virtsetpalette
   - virtgetpalette
   - virtpaletteset
   - virtpaletteput
   - virtpaletteinit
   - virtinit
   - virtreset
   - virtstore
   - virtrestore
   - virtline
   - virtscreencopy
   - virtwrite0400
   - virtwrite0800
   - virtwrite2000
   - virtwrite4000

*/

/* cache flags */
/* the lower 8 bits including cachepage are resevered for the switch counter */
#define cachechanged	0
#define cachepage	0x000001        /* (0 = page 1/ 1 = page 2) */
#define cachetext40	0x000100
#define cachetext40f	0x000200
#define cachetext40b	0x000400
#define cachetext40bf	0x000800
#define cachetext80	0x001000
#define cachelres40	0x002000
#define cachelres80	0x004000
#define cachehres	0x008000
#define cachehreso	0x010000
#define cachedhrs	0x020000

/* videomodes */
#define MODETEXT40	0
#define MODETEXT40BLUR	1
#define MODETEXT80	2
#define MODETEXT80BLUR	3
#define MODELRES	4
#define MODELRESMONO	5
#define MODEDLRS	6
#define MODEDLRSMONO	7
#define MODEHRES	8
#define MODEHRESRGB	9
#define MODEHRESBLUR	10
#define MODEHRESOFF	11
#define MODEHRESRGBOFF	12
#define MODEHRESBLUROFF	13
#define MODEHRESMONO	14
#define MODEDHRS	15
#define MODEDHRS140	16
#define MODEDHRSRGB	17
#define MODEDHRSMONO	18

static Charset       virtcharset;
static unsigned char virttextmode;
static unsigned char virthresmode;
static unsigned char virtdhrsmode;
static unsigned char virtmonochrome;
static unsigned char virtmonocol0;
static unsigned char virtmonocol1;
static unsigned int virtmodetop;
static unsigned int virtmodedown;
static unsigned int virttextpage;       /* 0x0400 or 0x0800 */
static unsigned int virthrespage;       /* 0x2000 or 0x4000 */
static unsigned int virtsplit;
unsigned int virtflashmode;
unsigned int virtflashmodeb;
unsigned int virtcachemode;
unsigned char virtcopy;                 /* screen update necessary? */
unsigned char virtgraphic;		/* softswitches $c050/51 */
unsigned char virtmixed;		/* softswitches $c052/53 */
unsigned char virtpage2;		/* softswitches $c054/55 */
unsigned char virthgr;			/* softswitches $c056/67 */
unsigned char virt80col;                /* 40col or 80col? */
unsigned char virtaltchar;              /* alternate charset? */
unsigned char virtiou;                  /* IOU active? */
unsigned char virtdhres;                /* Double Hires activated? */
unsigned char virtvideobyte;            /* byte put on the data bus by the graphic circuit */

static unsigned char colourtabe[8];
static unsigned char colourtabo[8];
static unsigned char colourtabe2[8];
static unsigned char colourtabo2[8];

unsigned char virtdouble;
unsigned char virtscanline;
static unsigned char virtpalette;

unsigned char virtfont[2048];
unsigned char virtfontflash[2048];
unsigned char virtfontalt[2048];
unsigned char *virtfontptr;
unsigned char *virtfontflashptr;

static unsigned int virtcache[192];
static unsigned int doublebits[256];

static unsigned int textline[0x400] = {
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      /* 0x0400 */
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     /* 0x0428 */
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,    /* 0x0450 */
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,      /* 0x0480 */
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,     /* 0x04a8 */
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,    /* 0x04d0 */
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,     /* 0x0500 */
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,     /* 0x0528 */
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,    /* 0x0550 */
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,     /* 0x0580 */
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,     /* 0x05a8 */
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,    /* 0x05d0 */
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,     /* 0x0600 */
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,     /* 0x0628 */
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,    /* 0x0650 */
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,     /* 0x0680 */
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,    /* 0x06a8 */
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,    /* 0x06d0 */
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,     /* 0x0700 */
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,    /* 0x0728 */
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,    /* 0x0750 */
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,     /* 0x0780 */
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,    /* 0x07a8 */
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,    /* 0x07d0 */
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                255,    255,    255,    255,    255,    255,    255,    255                     /* screenhole */
              };


static unsigned int textaddr[192] = {
                0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
                0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080,
                0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100, 0x0100,
                0x0180, 0x0180, 0x0180, 0x0180, 0x0180, 0x0180, 0x0180, 0x0180,
                0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200, 0x0200,
                0x0280, 0x0280, 0x0280, 0x0280, 0x0280, 0x0280, 0x0280, 0x0280,
                0x0300, 0x0300, 0x0300, 0x0300, 0x0300, 0x0300, 0x0300, 0x0300,
                0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380, 0x0380,
                0x0028, 0x0028, 0x0028, 0x0028, 0x0028, 0x0028, 0x0028, 0x0028,
                0x00a8, 0x00a8, 0x00a8, 0x00a8, 0x00a8, 0x00a8, 0x00a8, 0x00a8,
                0x0128, 0x0128, 0x0128, 0x0128, 0x0128, 0x0128, 0x0128, 0x0128,
                0x01a8, 0x01a8, 0x01a8, 0x01a8, 0x01a8, 0x01a8, 0x01a8, 0x01a8,
                0x0228, 0x0228, 0x0228, 0x0228, 0x0228, 0x0228, 0x0228, 0x0228,
                0x02a8, 0x02a8, 0x02a8, 0x02a8, 0x02a8, 0x02a8, 0x02a8, 0x02a8,
                0x0328, 0x0328, 0x0328, 0x0328, 0x0328, 0x0328, 0x0328, 0x0328,
                0x03a8, 0x03a8, 0x03a8, 0x03a8, 0x03a8, 0x03a8, 0x03a8, 0x03a8,
                0x0050, 0x0050, 0x0050, 0x0050, 0x0050, 0x0050, 0x0050, 0x0050,
                0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0, 0x00d0,
                0x0150, 0x0150, 0x0150, 0x0150, 0x0150, 0x0150, 0x0150, 0x0150,
                0x01d0, 0x01d0, 0x01d0, 0x01d0, 0x01d0, 0x01d0, 0x01d0, 0x01d0,
                0x0250, 0x0250, 0x0250, 0x0250, 0x0250, 0x0250, 0x0250, 0x0250,
                0x02d0, 0x02d0, 0x02d0, 0x02d0, 0x02d0, 0x02d0, 0x02d0, 0x02d0,
                0x0350, 0x0350, 0x0350, 0x0350, 0x0350, 0x0350, 0x0350, 0x0350,
                0x03d0, 0x03d0, 0x03d0, 0x03d0, 0x03d0, 0x03d0, 0x03d0, 0x03d0
                };


static unsigned int hresline[0x2000] = {
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,      /* 0x2000 */
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                0,      0,      0,      0,      0,      0,      0,      0,      0,      0,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,     /* 0x2028 */
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                64,     64,     64,     64,     64,     64,     64,     64,     64,     64,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,    /* 0x2050 */
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                128,    128,    128,    128,    128,    128,    128,    128,    128,    128,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,      /* 0x2080 */
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                8,      8,      8,      8,      8,      8,      8,      8,      8,      8,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,     /* 0x20a8 */
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                72,     72,     72,     72,     72,     72,     72,     72,     72,     72,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,    /* 0x20d0 */
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                136,    136,    136,    136,    136,    136,    136,    136,    136,    136,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,     /* 0x2100 */
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                16,     16,     16,     16,     16,     16,     16,     16,     16,     16,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,     /* 0x2128 */
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                80,     80,     80,     80,     80,     80,     80,     80,     80,     80,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,    /* 0x2150 */
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                144,    144,    144,    144,    144,    144,    144,    144,    144,    144,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,     /* 0x2180 */
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                24,     24,     24,     24,     24,     24,     24,     24,     24,     24,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,     /* 0x21a8 */
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                88,     88,     88,     88,     88,     88,     88,     88,     88,     88,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,    /* 0x21d0 */
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                152,    152,    152,    152,    152,    152,    152,    152,    152,    152,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,     /* 0x2200 */
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                32,     32,     32,     32,     32,     32,     32,     32,     32,     32,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,     /* 0x2228 */
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                96,     96,     96,     96,     96,     96,     96,     96,     96,     96,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,    /* 0x2250 */
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                160,    160,    160,    160,    160,    160,    160,    160,    160,    160,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,     /* 0x2280 */
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                40,     40,     40,     40,     40,     40,     40,     40,     40,     40,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,    /* 0x22a8 */
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                104,    104,    104,    104,    104,    104,    104,    104,    104,    104,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,    /* 0x22d0 */
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                168,    168,    168,    168,    168,    168,    168,    168,    168,    168,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,     /* 0x2300 */
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                48,     48,     48,     48,     48,     48,     48,     48,     48,     48,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,    /* 0x2328 */
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                112,    112,    112,    112,    112,    112,    112,    112,    112,    112,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,    /* 0x2350 */
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                176,    176,    176,    176,    176,    176,    176,    176,    176,    176,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,     /* 0x2380 */
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                56,     56,     56,     56,     56,     56,     56,     56,     56,     56,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,    /* 0x23a8 */
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                120,    120,    120,    120,    120,    120,    120,    120,    120,    120,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,    /* 0x23d0 */
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                184,    184,    184,    184,    184,    184,    184,    184,    184,    184,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                1,      1,      1,      1,      1,      1,      1,      1,      1,      1,      /* 0x2400 */
                1,      1,      1,      1,      1,      1,      1,      1,      1,      1,
                1,      1,      1,      1,      1,      1,      1,      1,      1,      1,
                1,      1,      1,      1,      1,      1,      1,      1,      1,      1,
                65,     65,     65,     65,     65,     65,     65,     65,     65,     65,     /* 0x2428 */
                65,     65,     65,     65,     65,     65,     65,     65,     65,     65,
                65,     65,     65,     65,     65,     65,     65,     65,     65,     65,
                65,     65,     65,     65,     65,     65,     65,     65,     65,     65,
                129,    129,    129,    129,    129,    129,    129,    129,    129,    129,    /* 0x2450 */
                129,    129,    129,    129,    129,    129,    129,    129,    129,    129,
                129,    129,    129,    129,    129,    129,    129,    129,    129,    129,
                129,    129,    129,    129,    129,    129,    129,    129,    129,    129,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                9,      9,      9,      9,      9,      9,      9,      9,      9,      9,      /* 0x2480 */
                9,      9,      9,      9,      9,      9,      9,      9,      9,      9,
                9,      9,      9,      9,      9,      9,      9,      9,      9,      9,
                9,      9,      9,      9,      9,      9,      9,      9,      9,      9,
                73,     73,     73,     73,     73,     73,     73,     73,     73,     73,     /* 0x24a8 */
                73,     73,     73,     73,     73,     73,     73,     73,     73,     73,
                73,     73,     73,     73,     73,     73,     73,     73,     73,     73,
                73,     73,     73,     73,     73,     73,     73,     73,     73,     73,
                137,    137,    137,    137,    137,    137,    137,    137,    137,    137,    /* 0x24d0 */
                137,    137,    137,    137,    137,    137,    137,    137,    137,    137,
                137,    137,    137,    137,    137,    137,    137,    137,    137,    137,
                137,    137,    137,    137,    137,    137,    137,    137,    137,    137,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                17,     17,     17,     17,     17,     17,     17,     17,     17,     17,     /* 0x2500 */
                17,     17,     17,     17,     17,     17,     17,     17,     17,     17,
                17,     17,     17,     17,     17,     17,     17,     17,     17,     17,
                17,     17,     17,     17,     17,     17,     17,     17,     17,     17,
                81,     81,     81,     81,     81,     81,     81,     81,     81,     81,     /* 0x2528 */
                81,     81,     81,     81,     81,     81,     81,     81,     81,     81,
                81,     81,     81,     81,     81,     81,     81,     81,     81,     81,
                81,     81,     81,     81,     81,     81,     81,     81,     81,     81,
                145,    145,    145,    145,    145,    145,    145,    145,    145,    145,    /* 0x2550 */
                145,    145,    145,    145,    145,    145,    145,    145,    145,    145,
                145,    145,    145,    145,    145,    145,    145,    145,    145,    145,
                145,    145,    145,    145,    145,    145,    145,    145,    145,    145,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                25,     25,     25,     25,     25,     25,     25,     25,     25,     25,     /* 0x2580 */
                25,     25,     25,     25,     25,     25,     25,     25,     25,     25,
                25,     25,     25,     25,     25,     25,     25,     25,     25,     25,
                25,     25,     25,     25,     25,     25,     25,     25,     25,     25,
                89,     89,     89,     89,     89,     89,     89,     89,     89,     89,     /* 0x25a8 */
                89,     89,     89,     89,     89,     89,     89,     89,     89,     89,
                89,     89,     89,     89,     89,     89,     89,     89,     89,     89,
                89,     89,     89,     89,     89,     89,     89,     89,     89,     89,
                153,    153,    153,    153,    153,    153,    153,    153,    153,    153,    /* 0x25d0 */
                153,    153,    153,    153,    153,    153,    153,    153,    153,    153,
                153,    152,    153,    153,    153,    153,    153,    153,    153,    153,
                153,    153,    153,    153,    153,    153,    153,    153,    153,    153,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                33,     33,     33,     33,     33,     33,     33,     33,     33,     33,     /* 0x2600 */
                33,     33,     33,     33,     33,     33,     33,     33,     33,     33,
                33,     33,     33,     33,     33,     33,     33,     33,     33,     33,
                33,     33,     33,     33,     33,     33,     33,     33,     33,     33,
                97,     97,     97,     97,     97,     97,     97,     97,     97,     97,     /* 0x2628 */
                97,     97,     97,     97,     97,     97,     97,     97,     97,     97,
                97,     97,     97,     97,     97,     97,     97,     97,     97,     97,
                97,     97,     97,     97,     97,     97,     97,     97,     97,     97,
                161,    161,    161,    161,    161,    161,    161,    161,    161,    161,    /* 0x2650 */
                161,    161,    161,    161,    161,    161,    161,    161,    161,    161,
                161,    161,    161,    161,    161,    161,    161,    161,    161,    161,
                161,    161,    161,    161,    161,    161,    161,    161,    161,    161,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                41,     41,     41,     41,     41,     41,     41,     41,     41,     41,     /* 0x2680 */
                41,     41,     41,     41,     41,     41,     41,     41,     41,     41,
                41,     41,     41,     41,     41,     41,     41,     41,     41,     41,
                41,     41,     41,     41,     41,     41,     41,     41,     41,     41,
                105,    105,    105,    105,    105,    105,    105,    105,    105,    105,    /* 0x26a8 */
                105,    105,    105,    105,    105,    105,    105,    105,    105,    105,
                105,    105,    105,    105,    105,    105,    105,    105,    105,    105,
                105,    105,    105,    105,    105,    105,    105,    105,    105,    105,
                169,    169,    169,    169,    169,    169,    169,    169,    169,    169,    /* 0x26d0 */
                169,    169,    169,    169,    169,    169,    169,    169,    169,    169,
                169,    169,    169,    169,    169,    169,    169,    169,    169,    169,
                169,    169,    169,    169,    169,    169,    169,    169,    169,    169,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                49,     49,     49,     49,     49,     49,     49,     49,     49,     49,     /* 0x2700 */
                49,     49,     49,     49,     49,     49,     49,     49,     49,     49,
                49,     49,     49,     49,     49,     49,     49,     49,     49,     49,
                49,     49,     49,     49,     49,     49,     49,     49,     49,     49,
                113,    113,    113,    113,    113,    113,    113,    113,    113,    113,    /* 0x2728 */
                113,    113,    113,    113,    113,    113,    113,    113,    113,    113,
                113,    113,    113,    113,    113,    113,    113,    113,    113,    113,
                113,    113,    113,    113,    113,    113,    113,    113,    113,    113,
                177,    177,    177,    177,    177,    177,    177,    177,    177,    177,    /* 0x2750 */
                177,    177,    177,    177,    177,    177,    177,    177,    177,    177,
                177,    177,    177,    177,    177,    177,    177,    177,    177,    177,
                177,    177,    177,    177,    177,    177,    177,    177,    177,    177,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                57,     57,     57,     57,     57,     57,     57,     57,     57,     57,     /* 0x2780 */
                57,     57,     57,     57,     57,     57,     57,     57,     57,     57,
                57,     57,     57,     57,     57,     57,     57,     57,     57,     57,
                57,     57,     57,     57,     57,     57,     57,     57,     57,     57,
                121,    121,    121,    121,    121,    121,    121,    121,    121,    121,    /* 0x27a8 */
                121,    121,    121,    121,    121,    121,    121,    121,    121,    121,
                121,    121,    121,    121,    121,    121,    121,    121,    121,    121,
                121,    121,    121,    121,    121,    121,    121,    121,    121,    121,
                185,    185,    185,    185,    185,    185,    185,    185,    185,    185,    /* 0x27d0 */
                185,    185,    185,    185,    185,    185,    185,    185,    185,    185,
                185,    185,    185,    185,    185,    185,    185,    185,    185,    185,
                185,    185,    185,    185,    185,    185,    185,    185,    185,    185,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                2,      2,      2,      2,      2,      2,      2,      2,      2,      2,      /* 0x2800 */
                2,      2,      2,      2,      2,      2,      2,      2,      2,      2,
                2,      2,      2,      2,      2,      2,      2,      2,      2,      2,
                2,      2,      2,      2,      2,      2,      2,      2,      2,      2,
                66,     66,     66,     66,     66,     66,     66,     66,     66,     66,     /* 0x2828 */
                66,     66,     66,     66,     66,     66,     66,     66,     66,     66,
                66,     66,     66,     66,     66,     66,     66,     66,     66,     66,
                66,     66,     66,     66,     66,     66,     66,     66,     66,     66,
                130,    130,    130,    130,    130,    130,    130,    130,    130,    130,    /* 0x2850 */
                130,    130,    130,    130,    130,    130,    130,    130,    130,    130,
                130,    130,    130,    130,    130,    130,    130,    130,    130,    130,
                130,    130,    130,    130,    130,    130,    130,    130,    130,    130,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                10,     10,     10,     10,     10,     10,     10,     10,     10,     10,     /* 0x2880 */
                10,     10,     10,     10,     10,     10,     10,     10,     10,     10,
                10,     10,     10,     10,     10,     10,     10,     10,     10,     10,
                10,     10,     10,     10,     10,     10,     10,     10,     10,     10,
                74,     74,     74,     74,     74,     74,     74,     74,     74,     74,     /* 0x28a8 */
                74,     74,     74,     74,     74,     74,     74,     74,     74,     74,
                74,     74,     74,     74,     74,     74,     74,     74,     74,     74,
                74,     74,     74,     74,     74,     74,     74,     74,     74,     74,
                138,    138,    138,    138,    138,    138,    138,    138,    138,    138,    /* 0x28d0 */
                138,    138,    138,    138,    138,    138,    138,    138,    138,    138,
                138,    138,    138,    138,    138,    138,    138,    138,    138,    138,
                138,    138,    138,    138,    138,    138,    138,    138,    138,    138,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                18,     18,     18,     18,     18,     18,     18,     18,     18,     18,     /* 0x2900 */
                18,     18,     18,     18,     18,     18,     18,     18,     18,     18,
                18,     18,     18,     18,     18,     18,     18,     18,     18,     18,
                18,     18,     18,     18,     18,     18,     18,     18,     18,     18,
                82,     82,     82,     82,     82,     82,     82,     82,     82,     82,     /* 0x2928 */
                82,     82,     82,     82,     82,     82,     82,     82,     82,     82,
                82,     82,     82,     82,     82,     82,     82,     82,     82,     82,
                82,     82,     82,     82,     82,     82,     82,     82,     82,     82,
                146,    146,    146,    146,    146,    146,    146,    146,    146,    146,    /* 0x2950 */
                146,    146,    146,    146,    146,    146,    146,    146,    146,    146,
                146,    146,    146,    146,    146,    146,    146,    146,    146,    146,
                146,    146,    146,    146,    146,    146,    146,    146,    146,    146,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                26,     26,     26,     26,     26,     26,     26,     26,     26,     26,     /* 0x2980 */
                26,     26,     26,     26,     26,     26,     26,     26,     26,     26,
                26,     26,     26,     26,     26,     26,     26,     26,     26,     26,
                26,     26,     26,     26,     26,     26,     26,     26,     26,     26,
                90,     90,     90,     90,     90,     90,     90,     90,     90,     90,     /* 0x29a8 */
                90,     90,     90,     90,     90,     90,     90,     90,     90,     90,
                90,     90,     90,     90,     90,     90,     90,     90,     90,     90,
                90,     90,     90,     90,     90,     90,     90,     90,     90,     90,
                154,    154,    154,    154,    154,    154,    154,    154,    154,    154,    /* 0x29d0 */
                154,    154,    154,    154,    154,    154,    154,    154,    154,    154,
                154,    154,    154,    154,    154,    154,    154,    154,    154,    154,
                154,    154,    154,    154,    154,    154,    154,    154,    154,    154,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                34,     34,     34,     34,     34,     34,     34,     34,     34,     34,     /* 0x2a00 */
                34,     34,     34,     34,     34,     34,     34,     34,     34,     34,
                34,     34,     34,     34,     34,     34,     34,     34,     34,     34,
                34,     34,     34,     34,     34,     34,     34,     34,     34,     34,
                98,     98,     98,     98,     98,     98,     98,     98,     98,     98,     /* 0x2a28 */
                98,     98,     98,     98,     98,     98,     98,     98,     98,     98,
                98,     98,     98,     98,     98,     98,     98,     98,     98,     98,
                98,     98,     98,     98,     98,     98,     98,     98,     98,     98,
                162,    162,    162,    162,    162,    162,    162,    162,    162,    162,    /* 0x2a50 */
                162,    162,    162,    162,    162,    162,    162,    162,    162,    162,
                162,    162,    162,    162,    162,    162,    162,    162,    162,    162,
                162,    162,    162,    162,    162,    162,    162,    162,    162,    162,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                42,     42,     42,     42,     42,     42,     42,     42,     42,     42,     /* 0x2a80 */
                42,     42,     42,     42,     42,     42,     42,     42,     42,     42,
                42,     42,     42,     42,     42,     42,     42,     42,     42,     42,
                42,     42,     42,     42,     42,     42,     42,     42,     42,     42,
                106,    106,    106,    106,    106,    106,    106,    106,    106,    106,    /* 0x2aa8 */
                106,    106,    106,    106,    106,    106,    106,    106,    106,    106,
                106,    106,    106,    106,    106,    106,    106,    106,    106,    106,
                106,    106,    106,    106,    106,    106,    106,    106,    106,    106,
                170,    170,    170,    170,    170,    170,    170,    170,    170,    170,    /* 0x2ad0 */
                170,    170,    170,    170,    170,    170,    170,    170,    170,    170,
                170,    170,    170,    170,    170,    170,    170,    170,    170,    170,
                170,    170,    170,    170,    170,    170,    170,    170,    170,    170,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                50,     50,     50,     50,     50,     50,     50,     50,     50,     50,     /* 0x2b00 */
                50,     50,     50,     50,     50,     50,     50,     50,     50,     50,
                50,     50,     50,     50,     50,     50,     50,     50,     50,     50,
                50,     50,     50,     50,     50,     50,     50,     50,     50,     50,
                114,    114,    114,    114,    114,    114,    114,    114,    114,    114,    /* 0x2b28 */
                114,    114,    114,    114,    114,    114,    114,    114,    114,    114,
                114,    114,    114,    114,    114,    114,    114,    114,    114,    114,
                114,    114,    114,    114,    114,    114,    114,    114,    114,    114,
                178,    178,    178,    178,    178,    178,    178,    178,    178,    178,    /* 0x2b50 */
                178,    178,    178,    178,    178,    178,    178,    178,    178,    178,
                178,    178,    178,    178,    178,    178,    178,    178,    178,    178,
                178,    178,    178,    178,    178,    178,    178,    178,    178,    178,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                58,     58,     58,     58,     58,     58,     58,     58,     58,     58,     /* 0x2b80 */
                58,     58,     58,     58,     58,     58,     58,     58,     58,     58,
                58,     58,     58,     58,     58,     58,     58,     58,     58,     58,
                58,     58,     58,     58,     58,     58,     58,     58,     58,     58,
                122,    122,    122,    122,    122,    122,    122,    122,    122,    122,    /* 0x2ba8 */
                122,    122,    122,    122,    122,    122,    122,    122,    122,    122,
                122,    122,    122,    122,    122,    122,    122,    122,    122,    122,
                122,    122,    122,    122,    122,    122,    122,    122,    122,    122,
                186,    186,    186,    186,    186,    186,    186,    186,    186,    186,    /* 0x2bd0 */
                186,    186,    186,    186,    186,    186,    186,    186,    186,    186,
                186,    186,    186,    186,    186,    186,    186,    186,    186,    186,
                186,    186,    186,    186,    186,    186,    186,    186,    186,    186,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                3,      3,      3,      3,      3,      3,      3,      3,      3,      3,      /* 0x2c00 */
                3,      3,      3,      3,      3,      3,      3,      3,      3,      3,
                3,      3,      3,      3,      3,      3,      3,      3,      3,      3,
                3,      3,      3,      3,      3,      3,      3,      3,      3,      3,
                67,     67,     67,     67,     67,     67,     67,     67,     67,     67,     /* 0x2c28 */
                67,     67,     67,     67,     67,     67,     67,     67,     67,     67,
                67,     67,     67,     67,     67,     67,     67,     67,     67,     67,
                67,     67,     67,     67,     67,     67,     67,     67,     67,     67,
                131,    131,    131,    131,    131,    131,    131,    131,    131,    131,    /* 0x2c50 */
                131,    131,    131,    131,    131,    131,    131,    131,    131,    131,
                131,    131,    131,    131,    131,    131,    131,    131,    131,    131,
                131,    131,    131,    131,    131,    131,    131,    131,    131,    131,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                11,     11,     11,     11,     11,     11,     11,     11,     11,     11,     /* 0x2c80 */
                11,     11,     11,     11,     11,     11,     11,     11,     11,     11,
                11,     11,     11,     11,     11,     11,     11,     11,     11,     11,
                11,     11,     11,     11,     11,     11,     11,     11,     11,     11,
                75,     75,     75,     75,     75,     75,     75,     75,     75,     75,     /* 0x2ca8 */
                75,     75,     75,     75,     75,     75,     75,     75,     75,     75,
                75,     75,     75,     75,     75,     75,     75,     75,     75,     75,
                75,     75,     75,     75,     75,     75,     75,     75,     75,     75,
                139,    139,    139,    139,    139,    139,    139,    139,    139,    139,    /* 0x2cd0 */
                139,    139,    139,    139,    139,    139,    139,    139,    139,    139,
                139,    139,    139,    139,    139,    139,    139,    139,    139,    139,
                139,    139,    139,    139,    139,    139,    139,    139,    139,    139,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                19,     19,     19,     19,     19,     19,     19,     19,     19,     19,     /* 0x2d00 */
                19,     19,     19,     19,     19,     19,     19,     19,     19,     19,
                19,     19,     19,     19,     19,     19,     19,     19,     19,     19,
                19,     19,     19,     19,     19,     19,     19,     19,     19,     19,
                83,     83,     83,     83,     83,     83,     83,     83,     83,     83,     /* 0x2d28 */
                83,     83,     83,     83,     83,     83,     83,     83,     83,     83,
                83,     83,     83,     83,     83,     83,     83,     83,     83,     83,
                83,     83,     83,     83,     83,     83,     83,     83,     83,     83,
                147,    147,    147,    147,    147,    147,    147,    147,    147,    147,    /* 0x2d50 */
                147,    147,    147,    147,    147,    147,    147,    147,    147,    147,
                147,    147,    147,    147,    147,    147,    147,    147,    147,    147,
                147,    147,    147,    147,    147,    147,    147,    147,    147,    147,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                27,     27,     27,     27,     27,     27,     27,     27,     27,     27,     /* 0x2d80 */
                27,     27,     27,     27,     27,     27,     27,     27,     27,     27,
                27,     27,     27,     27,     27,     27,     27,     27,     27,     27,
                27,     27,     27,     27,     27,     27,     27,     27,     27,     27,
                91,     91,     91,     91,     91,     91,     91,     91,     91,     91,     /* 0x2da8 */
                91,     91,     91,     91,     91,     91,     91,     91,     91,     91,
                91,     91,     91,     91,     91,     91,     91,     91,     91,     91,
                91,     91,     91,     91,     91,     91,     91,     91,     91,     91,
                155,    155,    155,    155,    155,    155,    155,    155,    155,    155,    /* 0x2dd0 */
                155,    155,    155,    155,    155,    155,    155,    155,    155,    155,
                155,    155,    155,    155,    155,    155,    155,    155,    155,    155,
                155,    155,    155,    155,    155,    155,    155,    155,    155,    155,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                35,     35,     35,     35,     35,     35,     35,     35,     35,     35,     /* 0x2e00 */
                35,     35,     35,     35,     35,     35,     35,     35,     35,     35,
                35,     35,     35,     35,     35,     35,     35,     35,     35,     35,
                35,     35,     35,     35,     35,     35,     35,     35,     35,     35,
                99,     99,     99,     99,     99,     99,     99,     99,     99,     99,     /* 0x2e28 */
                99,     99,     99,     99,     99,     99,     99,     99,     99,     99,
                99,     99,     99,     99,     99,     99,     99,     99,     99,     99,
                99,     99,     99,     99,     99,     99,     99,     99,     99,     99,
                163,    163,    163,    163,    163,    163,    163,    163,    163,    163,    /* 0x2e50 */
                163,    163,    163,    163,    163,    163,    163,    163,    163,    163,
                163,    163,    163,    163,    163,    163,    163,    163,    163,    163,
                163,    163,    163,    163,    163,    163,    163,    163,    163,    163,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                43,     43,     43,     43,     43,     43,     43,     43,     43,     43,     /* 0x2e80 */
                43,     43,     43,     43,     43,     43,     43,     43,     43,     43,
                43,     43,     43,     43,     43,     43,     43,     43,     43,     43,
                43,     43,     43,     43,     43,     43,     43,     43,     43,     43,
                107,    107,    107,    107,    107,    107,    107,    107,    107,    107,    /* 0x2ea8 */
                107,    107,    107,    107,    107,    107,    107,    107,    107,    107,
                107,    107,    107,    107,    107,    107,    107,    107,    107,    107,
                107,    107,    107,    107,    107,    107,    107,    107,    107,    107,
                171,    171,    171,    171,    171,    171,    171,    171,    171,    171,    /* 0x2ed0 */
                171,    171,    171,    171,    171,    171,    171,    171,    171,    171,
                171,    171,    171,    171,    171,    171,    171,    171,    171,    171,
                171,    171,    171,    171,    171,    171,    171,    171,    171,    171,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                51,     51,     51,     51,     51,     51,     51,     51,     51,     51,     /* 0x2f00 */
                51,     51,     51,     51,     51,     51,     51,     51,     51,     51,
                51,     51,     51,     51,     51,     51,     51,     51,     51,     51,
                51,     51,     51,     51,     51,     51,     51,     51,     51,     51,
                115,    115,    115,    115,    115,    115,    115,    115,    115,    115,    /* 0x2f28 */
                115,    115,    115,    115,    115,    115,    115,    115,    115,    115,
                115,    115,    115,    115,    115,    115,    115,    115,    115,    115,
                115,    115,    115,    115,    115,    115,    115,    115,    115,    115,
                179,    179,    179,    179,    179,    179,    179,    179,    179,    179,    /* 0x2f50 */
                179,    179,    179,    179,    179,    179,    179,    179,    179,    179,
                179,    179,    179,    179,    179,    179,    179,    179,    179,    179,
                179,    179,    179,    179,    179,    179,    179,    179,    179,    179,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                59,     59,     59,     59,     59,     59,     59,     59,     59,     59,     /* 0x2f80 */
                59,     59,     59,     59,     59,     59,     59,     59,     59,     59,
                59,     59,     59,     59,     59,     59,     59,     59,     59,     59,
                59,     59,     59,     59,     59,     59,     59,     59,     59,     59,
                123,    123,    123,    123,    123,    123,    123,    123,    123,    123,    /* 0x2fa8 */
                123,    123,    123,    123,    123,    123,    123,    123,    123,    123,
                123,    123,    123,    123,    123,    123,    123,    123,    123,    123,
                123,    123,    123,    123,    123,    123,    123,    123,    123,    123,
                187,    187,    187,    187,    187,    187,    187,    187,    187,    187,    /* 0x2fd0 */
                187,    187,    187,    187,    187,    187,    187,    187,    187,    187,
                187,    187,    187,    187,    187,    187,    187,    187,    187,    187,
                187,    187,    187,    187,    187,    187,    187,    187,    187,    187,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                4,      4,      4,      4,      4,      4,      4,      4,      4,      4,      /* 0x3000 */
                4,      4,      4,      4,      4,      4,      4,      4,      4,      4,
                4,      4,      4,      4,      4,      4,      4,      4,      4,      4,
                4,      4,      4,      4,      4,      4,      4,      4,      4,      4,
                68,     68,     68,     68,     68,     68,     68,     68,     68,     68,     /* 0x3028 */
                68,     68,     68,     68,     68,     68,     68,     68,     68,     68,
                68,     68,     68,     68,     68,     68,     68,     68,     68,     68,
                68,     68,     68,     68,     68,     68,     68,     68,     68,     68,
                132,    132,    132,    132,    132,    132,    132,    132,    132,    132,    /* 0x3050 */
                132,    132,    132,    132,    132,    132,    132,    132,    132,    132,
                132,    132,    132,    132,    132,    132,    132,    132,    132,    132,
                132,    132,    132,    132,    132,    132,    132,    132,    132,    132,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                12,     12,     12,     12,     12,     12,     12,     12,     12,     12,     /* 0x3080 */
                12,     12,     12,     12,     12,     12,     12,     12,     12,     12,
                12,     12,     12,     12,     12,     12,     12,     12,     12,     12,
                12,     12,     12,     12,     12,     12,     12,     12,     12,     12,
                76,     76,     76,     76,     76,     76,     76,     76,     76,     76,     /* 0x30a8 */
                76,     76,     76,     76,     76,     76,     76,     76,     76,     76,
                76,     76,     76,     76,     76,     76,     76,     76,     76,     76,
                76,     76,     76,     76,     76,     76,     76,     76,     76,     76,
                140,    140,    140,    140,    140,    140,    140,    140,    140,    140,    /* 0x30d0 */
                140,    140,    140,    140,    140,    140,    140,    140,    140,    140,
                140,    140,    140,    140,    140,    140,    140,    140,    140,    140,
                140,    140,    140,    140,    140,    140,    140,    140,    140,    140,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                20,     20,     20,     20,     20,     20,     20,     20,     20,     20,     /* 0x3100 */
                20,     20,     20,     20,     20,     20,     20,     20,     20,     20,
                20,     20,     20,     20,     20,     20,     20,     20,     20,     20,
                20,     20,     20,     20,     20,     20,     20,     20,     20,     20,
                84,     84,     84,     84,     84,     84,     84,     84,     84,     84,     /* 0x3128 */
                84,     84,     84,     84,     84,     84,     84,     84,     84,     84,
                84,     84,     84,     84,     84,     84,     84,     84,     84,     84,
                84,     84,     84,     84,     84,     84,     84,     84,     84,     84,
                148,    148,    148,    148,    148,    148,    148,    148,    148,    148,    /* 0x3150 */
                148,    148,    148,    148,    148,    148,    148,    148,    148,    148,
                148,    148,    148,    148,    148,    148,    148,    148,    148,    148,
                148,    148,    148,    148,    148,    148,    148,    148,    148,    148,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                28,     28,     28,     28,     28,     28,     28,     28,     28,     28,     /* 0x3180 */
                28,     28,     28,     28,     28,     28,     28,     28,     28,     28,
                28,     28,     28,     28,     28,     28,     28,     28,     28,     28,
                28,     28,     28,     28,     28,     28,     28,     28,     28,     28,
                92,     92,     92,     92,     92,     92,     92,     92,     92,     92,     /* 0x31a8 */
                92,     92,     92,     92,     92,     92,     92,     92,     92,     92,
                92,     92,     92,     92,     92,     92,     92,     92,     92,     92,
                92,     92,     92,     92,     92,     92,     92,     92,     92,     92,
                156,    156,    156,    156,    156,    156,    156,    156,    156,    156,    /* 0x31d0 */
                156,    156,    156,    156,    156,    156,    156,    156,    156,    156,
                156,    156,    156,    156,    156,    156,    156,    156,    156,    156,
                156,    156,    156,    156,    156,    156,    156,    156,    156,    156,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                36,     36,     36,     36,     36,     36,     36,     36,     36,     36,     /* 0x3200 */
                36,     36,     36,     36,     36,     36,     36,     36,     36,     36,
                36,     36,     36,     36,     36,     36,     36,     36,     36,     36,
                36,     36,     36,     36,     36,     36,     36,     36,     36,     36,
                100,    100,    100,    100,    100,    100,    100,    100,    100,    100,    /* 0x3228 */
                100,    100,    100,    100,    100,    100,    100,    100,    100,    100,
                100,    100,    100,    100,    100,    100,    100,    100,    100,    100,
                100,    100,    100,    100,    100,    100,    100,    100,    100,    100,
                164,    164,    164,    164,    164,    164,    164,    164,    164,    164,    /* 0x3250 */
                164,    164,    164,    164,    164,    164,    164,    164,    164,    164,
                164,    164,    164,    164,    164,    164,    164,    164,    164,    164,
                164,    164,    164,    164,    164,    164,    164,    164,    164,    164,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                44,     44,     44,     44,     44,     44,     44,     44,     44,     44,     /* 0x3280 */
                44,     44,     44,     44,     44,     44,     44,     44,     44,     44,
                44,     44,     44,     44,     44,     44,     44,     44,     44,     44,
                44,     44,     44,     44,     44,     44,     44,     44,     44,     44,
                108,    108,    108,    108,    108,    108,    108,    108,    108,    108,    /* 0x32a8 */
                108,    108,    108,    108,    108,    108,    108,    108,    108,    108,
                108,    108,    108,    108,    108,    108,    108,    108,    108,    108,
                108,    108,    108,    108,    108,    108,    108,    108,    108,    108,
                172,    172,    172,    172,    172,    172,    172,    172,    172,    172,    /* 0x32d0 */
                172,    172,    172,    172,    172,    172,    172,    172,    172,    172,
                172,    172,    172,    172,    172,    172,    172,    172,    172,    172,
                172,    172,    172,    172,    172,    172,    172,    172,    172,    172,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                52,     52,     52,     52,     52,     52,     52,     52,     52,     52,     /* 0x3300 */
                52,     52,     52,     52,     52,     52,     52,     52,     52,     52,
                52,     52,     52,     52,     52,     52,     52,     52,     52,     52,
                52,     52,     52,     52,     52,     52,     52,     52,     52,     52,
                116,    116,    116,    116,    116,    116,    116,    116,    116,    116,    /* 0x3328 */
                116,    116,    116,    116,    116,    116,    116,    116,    116,    116,
                116,    116,    116,    116,    116,    116,    116,    116,    116,    116,
                116,    116,    116,    116,    116,    116,    116,    116,    116,    116,
                180,    180,    180,    180,    180,    180,    180,    180,    180,    180,    /* 0x3350 */
                180,    180,    180,    180,    180,    180,    180,    180,    180,    180,
                180,    180,    180,    180,    180,    180,    180,    180,    180,    180,
                180,    180,    180,    180,    180,    180,    180,    180,    180,    180,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                60,     60,     60,     60,     60,     60,     60,     60,     60,     60,     /* 0x3380 */
                60,     60,     60,     60,     60,     60,     60,     60,     60,     60,
                60,     60,     60,     60,     60,     60,     60,     60,     60,     60,
                60,     60,     60,     60,     60,     60,     60,     60,     60,     60,
                124,    124,    124,    124,    124,    124,    124,    124,    124,    124,    /* 0x33a8 */
                124,    124,    124,    124,    124,    124,    124,    124,    124,    124,
                124,    124,    124,    124,    124,    124,    124,    124,    124,    124,
                124,    124,    124,    124,    124,    124,    124,    124,    124,    124,
                188,    188,    188,    188,    188,    188,    188,    188,    188,    188,    /* 0x33d0 */
                188,    188,    188,    188,    188,    188,    188,    188,    188,    188,
                188,    188,    188,    188,    188,    188,    188,    188,    188,    188,
                188,    188,    188,    188,    188,    188,    188,    188,    188,    188,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                5,      5,      5,      5,      5,      5,      5,      5,      5,      5,      /* 0x3400 */
                5,      5,      5,      5,      5,      5,      5,      5,      5,      5,
                5,      5,      5,      5,      5,      5,      5,      5,      5,      5,
                5,      5,      5,      5,      5,      5,      5,      5,      5,      5,
                69,     69,     69,     69,     69,     69,     69,     69,     69,     69,     /* 0x3428 */
                69,     69,     69,     69,     69,     69,     69,     69,     69,     69,
                69,     69,     69,     69,     69,     69,     69,     69,     69,     69,
                69,     69,     69,     69,     69,     69,     69,     69,     69,     69,
                133,    133,    133,    133,    133,    133,    133,    133,    133,    133,    /* 0x3450 */
                133,    133,    133,    133,    133,    133,    133,    133,    133,    133,
                133,    133,    133,    133,    133,    133,    133,    133,    133,    133,
                133,    133,    133,    133,    133,    133,    133,    133,    133,    133,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                13,     13,     13,     13,     13,     13,     13,     13,     13,     13,     /* 0x3480 */
                13,     13,     13,     13,     13,     13,     13,     13,     13,     13,
                13,     13,     13,     13,     13,     13,     13,     13,     13,     13,
                13,     13,     13,     13,     13,     13,     13,     13,     13,     13,
                77,     77,     77,     77,     77,     77,     77,     77,     77,     77,     /* 0x34a8 */
                77,     77,     77,     77,     77,     77,     77,     77,     77,     77,
                77,     77,     77,     77,     77,     77,     77,     77,     77,     77,
                77,     77,     77,     77,     77,     77,     77,     77,     77,     77,
                141,    141,    141,    141,    141,    141,    141,    141,    141,    141,    /* 0x34d0 */
                141,    141,    141,    141,    141,    141,    141,    141,    141,    141,
                141,    141,    141,    141,    141,    141,    141,    141,    141,    141,
                141,    141,    141,    141,    141,    141,    141,    141,    141,    141,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                21,     21,     21,     21,     21,     21,     21,     21,     21,     21,     /* 0x3500 */
                21,     21,     21,     21,     21,     21,     21,     21,     21,     21,
                21,     21,     21,     21,     21,     21,     21,     21,     21,     21,
                21,     21,     21,     21,     21,     21,     21,     21,     21,     21,
                85,     85,     85,     85,     85,     85,     85,     85,     85,     85,     /* 0x3528 */
                85,     85,     85,     85,     85,     85,     85,     85,     85,     85,
                85,     85,     85,     85,     85,     85,     85,     85,     85,     85,
                85,     85,     85,     85,     85,     85,     85,     85,     85,     85,
                149,    149,    149,    149,    149,    149,    149,    149,    149,    149,    /* 0x3550 */
                149,    149,    149,    149,    149,    149,    149,    149,    149,    149,
                149,    149,    149,    149,    149,    149,    149,    149,    149,    149,
                149,    149,    149,    149,    149,    149,    149,    149,    149,    149,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                29,     29,     29,     29,     29,     29,     29,     29,     29,     29,     /* 0x3580 */
                29,     29,     29,     29,     29,     29,     29,     29,     29,     29,
                29,     29,     29,     29,     29,     29,     29,     29,     29,     29,
                29,     29,     29,     29,     29,     29,     29,     29,     29,     29,
                93,     93,     93,     93,     93,     93,     93,     93,     93,     93,     /* 0x35a8 */
                93,     93,     93,     93,     93,     93,     93,     93,     93,     93,
                93,     93,     93,     93,     93,     93,     93,     93,     93,     93,
                93,     93,     93,     93,     93,     93,     93,     93,     93,     93,
                157,    157,    157,    157,    157,    157,    157,    157,    157,    157,    /* 0x35d0 */
                157,    157,    157,    157,    157,    157,    157,    157,    157,    157,
                157,    157,    157,    157,    157,    157,    157,    157,    157,    157,
                157,    157,    157,    157,    157,    157,    157,    157,    157,    157,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                37,     37,     37,     37,     37,     37,     37,     37,     37,     37,     /* 0x3600 */
                37,     37,     37,     37,     37,     37,     37,     37,     37,     37,
                37,     37,     37,     37,     37,     37,     37,     37,     37,     37,
                37,     37,     37,     37,     37,     37,     37,     37,     37,     37,
                101,    101,    101,    101,    101,    101,    101,    101,    101,    101,    /* 0x3628 */
                101,    101,    101,    101,    101,    101,    101,    101,    101,    101,
                101,    101,    101,    101,    101,    101,    101,    101,    101,    101,
                101,    101,    101,    101,    101,    101,    101,    101,    101,    101,
                165,    165,    165,    165,    165,    165,    165,    165,    165,    165,    /* 0x3650 */
                165,    165,    165,    165,    165,    165,    165,    165,    165,    165,
                165,    165,    165,    165,    165,    165,    165,    165,    165,    165,
                165,    165,    165,    165,    165,    165,    165,    165,    165,    165,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                45,     45,     45,     45,     45,     45,     45,     45,     45,     45,     /* 0x3680 */
                45,     45,     45,     45,     45,     45,     45,     45,     45,     45,
                45,     45,     45,     45,     45,     45,     45,     45,     45,     45,
                45,     45,     45,     45,     45,     45,     45,     45,     45,     45,
                109,    109,    109,    109,    109,    109,    109,    109,    109,    109,    /* 0x36a8 */
                109,    109,    109,    109,    109,    109,    109,    109,    109,    109,
                109,    109,    109,    109,    109,    109,    109,    109,    109,    109,
                109,    109,    109,    109,    109,    109,    109,    109,    109,    109,
                173,    173,    173,    173,    173,    173,    173,    173,    173,    173,    /* 0x36d0 */
                173,    173,    173,    173,    173,    173,    173,    173,    173,    173,
                173,    173,    173,    173,    173,    173,    173,    173,    173,    173,
                173,    173,    173,    173,    173,    173,    173,    173,    173,    173,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                53,     53,     53,     53,     53,     53,     53,     53,     53,     53,     /* 0x3700 */
                53,     53,     53,     53,     53,     53,     53,     53,     53,     53,
                53,     53,     53,     53,     53,     53,     53,     53,     53,     53,
                53,     53,     53,     53,     53,     53,     53,     53,     53,     53,
                117,    117,    117,    117,    117,    117,    117,    117,    117,    117,    /* 0x3728 */
                117,    117,    117,    117,    117,    117,    117,    117,    117,    117,
                117,    117,    117,    117,    117,    117,    117,    117,    117,    117,
                117,    117,    117,    117,    117,    117,    117,    117,    117,    117,
                181,    181,    181,    181,    181,    181,    181,    181,    181,    181,    /* 0x3750 */
                181,    181,    181,    181,    181,    181,    181,    181,    181,    181,
                181,    181,    181,    181,    181,    181,    181,    181,    181,    181,
                181,    181,    181,    181,    181,    181,    181,    181,    181,    181,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                61,     61,     61,     61,     61,     61,     61,     61,     61,     61,     /* 0x3780 */
                61,     61,     61,     61,     61,     61,     61,     61,     61,     61,
                61,     61,     61,     61,     61,     61,     61,     61,     61,     61,
                61,     61,     61,     61,     61,     61,     61,     61,     61,     61,
                125,    125,    125,    125,    125,    125,    125,    125,    125,    125,    /* 0x37a8 */
                125,    125,    125,    125,    125,    125,    125,    125,    125,    125,
                125,    125,    125,    125,    125,    125,    125,    125,    125,    125,
                125,    125,    125,    125,    125,    125,    125,    125,    125,    125,
                189,    189,    189,    189,    189,    189,    189,    189,    189,    189,    /* 0x37d0 */
                189,    189,    189,    189,    189,    189,    189,    189,    189,    189,
                189,    189,    189,    189,    189,    189,    189,    189,    189,    189,
                189,    189,    189,    189,    189,    189,    189,    189,    189,    189,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */

                6,      6,      6,      6,      6,      6,      6,      6,      6,      6,      /* 0x3800 */
                6,      6,      6,      6,      6,      6,      6,      6,      6,      6,
                6,      6,      6,      6,      6,      6,      6,      6,      6,      6,
                6,      6,      6,      6,      6,      6,      6,      6,      6,      6,
                70,     70,     70,     70,     70,     70,     70,     70,     70,     70,     /* 0x3828 */
                70,     70,     70,     70,     70,     70,     70,     70,     70,     70,
                70,     70,     70,     70,     70,     70,     70,     70,     70,     70,
                70,     70,     70,     70,     70,     70,     70,     70,     70,     70,
                134,    134,    134,    134,    134,    134,    134,    134,    134,    134,    /* 0x3850 */
                134,    134,    134,    134,    134,    134,    134,    134,    134,    134,
                134,    134,    134,    134,    134,    134,    134,    134,    134,    134,
                134,    134,    134,    134,    134,    134,    134,    134,    134,    134,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                14,     14,     14,     14,     14,     14,     14,     14,     14,     14,     /* 0x3880 */
                14,     14,     14,     14,     14,     14,     14,     14,     14,     14,
                14,     14,     14,     14,     14,     14,     14,     14,     14,     14,
                14,     14,     14,     14,     14,     14,     14,     14,     14,     14,
                78,     78,     78,     78,     78,     78,     78,     78,     78,     78,     /* 0x38a8 */
                78,     78,     78,     78,     78,     78,     78,     78,     78,     78,
                78,     78,     78,     78,     78,     78,     78,     78,     78,     78,
                78,     78,     78,     78,     78,     78,     78,     78,     78,     78,
                142,    142,    142,    142,    142,    142,    142,    142,    142,    142,    /* 0x38d0 */
                142,    142,    142,    142,    142,    142,    142,    142,    142,    142,
                142,    142,    142,    142,    142,    142,    142,    142,    142,    142,
                142,    142,    142,    142,    142,    142,    142,    142,    142,    142,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                22,     22,     22,     22,     22,     22,     22,     22,     22,     22,     /* 0x3900 */
                22,     22,     22,     22,     22,     22,     22,     22,     22,     22,
                22,     22,     22,     22,     22,     22,     22,     22,     22,     22,
                22,     22,     22,     22,     22,     22,     22,     22,     22,     22,
                86,     86,     86,     86,     86,     86,     86,     86,     86,     86,     /* 0x3928 */
                86,     86,     86,     86,     86,     86,     86,     86,     86,     86,
                86,     86,     86,     86,     86,     86,     86,     86,     86,     86,
                86,     86,     86,     86,     86,     86,     86,     86,     86,     86,
                150,    150,    150,    150,    150,    150,    150,    150,    150,    150,    /* 0x3950 */
                150,    150,    150,    150,    150,    150,    150,    150,    150,    150,
                150,    150,    150,    150,    150,    150,    150,    150,    150,    150,
                150,    150,    150,    150,    150,    150,    150,    150,    150,    150,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                30,     30,     30,     30,     30,     30,     30,     30,     30,     30,     /* 0x3980 */
                30,     30,     30,     30,     30,     30,     30,     30,     30,     30,
                30,     30,     30,     30,     30,     30,     30,     30,     30,     30,
                30,     30,     30,     30,     30,     30,     30,     30,     30,     30,
                94,     94,     94,     94,     94,     94,     94,     94,     94,     94,     /* 0x39a8 */
                94,     94,     94,     94,     94,     94,     94,     94,     94,     94,
                94,     94,     94,     94,     94,     94,     94,     94,     94,     94,
                94,     94,     94,     94,     94,     94,     94,     94,     94,     94,
                158,    158,    158,    158,    158,    158,    158,    158,    158,    158,    /* 0x39d0 */
                158,    158,    158,    158,    158,    158,    158,    158,    158,    158,
                158,    158,    158,    158,    158,    158,    158,    158,    158,    158,
                158,    158,    158,    158,    158,    158,    158,    158,    158,    158,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                38,     38,     38,     38,     38,     38,     38,     38,     38,     38,     /* 0x3a00 */
                38,     38,     38,     38,     38,     38,     38,     38,     38,     38,
                38,     38,     38,     38,     38,     38,     38,     38,     38,     38,
                38,     38,     38,     38,     38,     38,     38,     38,     38,     38,
                102,    102,    102,    102,    102,    102,    102,    102,    102,    102,    /* 0x3a28 */
                102,    102,    102,    102,    102,    102,    102,    102,    102,    102,
                102,    102,    102,    102,    102,    102,    102,    102,    102,    102,
                102,    102,    102,    102,    102,    102,    102,    102,    102,    102,
                166,    166,    166,    166,    166,    166,    166,    166,    166,    166,    /* 0x3a50 */
                166,    166,    166,    166,    166,    166,    166,    166,    166,    166,
                166,    166,    166,    166,    166,    166,    166,    166,    166,    166,
                166,    166,    166,    166,    166,    166,    166,    166,    166,    166,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                46,     46,     46,     46,     46,     46,     46,     46,     46,     46,     /* 0x3a80 */
                46,     46,     46,     46,     46,     46,     46,     46,     46,     46,
                46,     46,     46,     46,     46,     46,     46,     46,     46,     46,
                46,     46,     46,     46,     46,     46,     46,     46,     46,     46,
                110,    110,    110,    110,    110,    110,    110,    110,    110,    110,    /* 0x3aa8 */
                110,    110,    110,    110,    110,    110,    110,    110,    110,    110,
                110,    110,    110,    110,    110,    110,    110,    110,    110,    110,
                110,    110,    110,    110,    110,    110,    110,    110,    110,    110,
                174,    174,    174,    174,    174,    174,    174,    174,    174,    174,    /* 0x3ad0 */
                174,    174,    174,    174,    174,    174,    174,    174,    174,    174,
                174,    174,    174,    174,    174,    174,    174,    174,    174,    174,
                174,    174,    174,    174,    174,    174,    174,    174,    174,    174,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                54,     54,     54,     54,     54,     54,     54,     54,     54,     54,     /* 0x3b00 */
                54,     54,     54,     54,     54,     54,     54,     54,     54,     54,
                54,     54,     54,     54,     54,     54,     54,     54,     54,     54,
                54,     54,     54,     54,     54,     54,     54,     54,     54,     54,
                118,    118,    118,    118,    118,    118,    118,    118,    118,    118,    /* 0x3b28 */
                118,    118,    118,    118,    118,    118,    118,    118,    118,    118,
                118,    118,    118,    118,    118,    118,    118,    118,    118,    118,
                118,    118,    118,    118,    118,    118,    118,    118,    118,    118,
                182,    182,    182,    182,    182,    182,    182,    182,    182,    182,    /* 0x3b50 */
                182,    182,    182,    182,    182,    182,    182,    182,    182,    182,
                182,    182,    182,    182,    182,    182,    182,    182,    182,    182,
                182,    182,    182,    182,    182,    182,    182,    182,    182,    182,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                62,     62,     62,     62,     62,     62,     62,     62,     62,     62,     /* 0x3b80 */
                62,     62,     62,     62,     62,     62,     62,     62,     62,     62,
                62,     62,     62,     62,     62,     62,     62,     62,     62,     62,
                62,     62,     62,     62,     62,     62,     62,     62,     62,     62,
                126,    126,    126,    126,    126,    126,    126,    126,    126,    126,    /* 0x3ba8 */
                126,    126,    126,    126,    126,    126,    126,    126,    126,    126,
                126,    126,    126,    126,    126,    126,    126,    126,    126,    126,
                126,    126,    126,    126,    126,    126,    126,    126,    126,    126,
                190,    190,    190,    190,    190,    190,    190,    190,    190,    190,    /* 0x3bd0 */
                190,    190,    190,    190,    190,    190,    190,    190,    190,    190,
                190,    190,    190,    190,    190,    190,    190,    190,    190,    190,
                190,    190,    190,    190,    190,    190,    190,    190,    190,    190,
                255,    255,    255,    255,    255,    255,    255,    255,            /* screenhole */

                7,      7,      7,      7,      7,      7,      7,      7,      7,      7,      /* 0x3c00 */
                7,      7,      7,      7,      7,      7,      7,      7,      7,      7,
                7,      7,      7,      7,      7,      7,      7,      7,      7,      7,
                7,      7,      7,      7,      7,      7,      7,      7,      7,      7,
                71,     71,     71,     71,     71,     71,     71,     71,     71,     71,     /* 0x3c28 */
                71,     71,     71,     71,     71,     71,     71,     71,     71,     71,
                71,     71,     71,     71,     71,     71,     71,     71,     71,     71,
                71,     71,     71,     71,     71,     71,     71,     71,     71,     71,
                135,    135,    135,    135,    135,    135,    135,    135,    135,    135,    /* 0x3c50 */
                135,    135,    135,    135,    135,    135,    135,    135,    135,    135,
                135,    135,    135,    135,    135,    135,    135,    135,    135,    135,
                135,    135,    135,    135,    135,    135,    135,    135,    135,    135,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                15,     15,     15,     15,     15,     15,     15,     15,     15,     15,     /* 0x3c80 */
                15,     15,     15,     15,     15,     15,     15,     15,     15,     15,
                15,     15,     15,     15,     15,     15,     15,     15,     15,     15,
                15,     15,     15,     15,     15,     15,     15,     15,     15,     15,
                79,     79,     79,     79,     79,     79,     79,     79,     79,     79,     /* 0x3ca8 */
                79,     79,     79,     79,     79,     79,     79,     79,     79,     79,
                79,     79,     79,     79,     79,     79,     79,     79,     79,     79,
                79,     79,     79,     79,     79,     79,     79,     79,     79,     79,
                143,    143,    143,    143,    143,    143,    143,    143,    143,    143,    /* 0x3cd0 */
                143,    143,    143,    143,    143,    143,    143,    143,    143,    143,
                143,    143,    143,    143,    143,    143,    143,    143,    143,    143,
                143,    143,    143,    143,    143,    143,    143,    143,    143,    143,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                23,     23,     23,     23,     23,     23,     23,     23,     23,     23,     /* 0x3d00 */
                23,     23,     23,     23,     23,     23,     23,     23,     23,     23,
                23,     23,     23,     23,     23,     23,     23,     23,     23,     23,
                23,     23,     23,     23,     23,     23,     23,     23,     23,     23,
                87,     87,     87,     87,     87,     87,     87,     87,     87,     87,     /* 0x3d28 */
                87,     87,     87,     87,     87,     87,     87,     87,     87,     87,
                87,     87,     87,     87,     87,     87,     87,     87,     87,     87,
                87,     87,     87,     87,     87,     87,     87,     87,     87,     87,
                151,    151,    151,    151,    151,    151,    151,    151,    151,    151,    /* 0x3d50 */
                151,    151,    151,    151,    151,    151,    151,    151,    151,    151,
                151,    151,    151,    151,    151,    151,    151,    151,    151,    151,
                151,    151,    151,    151,    151,    151,    151,    151,    151,    151,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                31,     31,     31,     31,     31,     31,     31,     31,     31,     31,     /* 0x3d80 */
                31,     31,     31,     31,     31,     31,     31,     31,     31,     31,
                31,     31,     31,     31,     31,     31,     31,     31,     31,     31,
                31,     31,     31,     31,     31,     31,     31,     31,     31,     31,
                95,     95,     95,     95,     95,     95,     95,     95,     95,     95,     /* 0x3da8 */
                95,     95,     95,     95,     95,     95,     95,     95,     95,     95,
                95,     95,     95,     95,     95,     95,     95,     95,     95,     95,
                95,     95,     95,     95,     95,     95,     95,     95,     95,     95,
                159,    159,    159,    159,    159,    159,    159,    159,    159,    159,    /* 0x3dd0 */
                159,    159,    159,    159,    159,    159,    159,    159,    159,    159,
                159,    159,    159,    159,    159,    159,    159,    159,    159,    159,
                159,    159,    159,    159,    159,    159,    159,    159,    159,    159,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                39,     39,     39,     39,     39,     39,     39,     39,     39,     39,     /* 0x3e00 */
                39,     39,     39,     39,     39,     39,     39,     39,     39,     39,
                39,     39,     39,     39,     39,     39,     39,     39,     39,     39,
                39,     39,     39,     39,     39,     39,     39,     39,     39,     39,
                103,    103,    103,    103,    103,    103,    103,    103,    103,    103,    /* 0x3e28 */
                103,    103,    103,    103,    103,    103,    103,    103,    103,    103,
                103,    103,    103,    103,    103,    103,    103,    103,    103,    103,
                103,    103,    103,    103,    103,    103,    103,    103,    103,    103,
                167,    167,    167,    167,    167,    167,    167,    167,    167,    167,    /* 0x3e50 */
                167,    167,    167,    167,    167,    167,    167,    167,    167,    167,
                167,    167,    167,    167,    167,    167,    167,    167,    167,    167,
                167,    167,    167,    167,    167,    167,    167,    167,    167,    167,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                47,     47,     47,     47,     47,     47,     47,     47,     47,     47,     /* 0x3e80 */
                47,     47,     47,     47,     47,     47,     47,     47,     47,     47,
                47,     47,     47,     47,     47,     47,     47,     47,     47,     47,
                47,     47,     47,     47,     47,     47,     47,     47,     47,     47,
                111,    111,    111,    111,    111,    111,    111,    111,    111,    111,    /* 0x3ea8 */
                111,    111,    111,    111,    111,    111,    111,    111,    111,    111,
                111,    111,    111,    111,    111,    111,    111,    111,    111,    111,
                111,    111,    111,    111,    111,    111,    111,    111,    111,    111,
                175,    175,    175,    175,    175,    175,    175,    175,    175,    175,    /* 0x3ed0 */
                175,    175,    175,    175,    175,    175,    175,    175,    175,    175,
                175,    175,    175,    175,    175,    175,    175,    175,    175,    175,
                175,    175,    175,    175,    175,    175,    175,    175,    175,    175,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                55,     55,     55,     55,     55,     55,     55,     55,     55,     55,     /* 0x3f00 */
                55,     55,     55,     55,     55,     55,     55,     55,     55,     55,
                55,     55,     55,     55,     55,     55,     55,     55,     55,     55,
                55,     55,     55,     55,     55,     55,     55,     55,     55,     55,
                119,    119,    119,    119,    119,    119,    119,    119,    119,    119,    /* 0x3f28 */
                119,    119,    119,    119,    119,    119,    119,    119,    119,    119,
                119,    119,    119,    119,    119,    119,    119,    119,    119,    119,
                119,    119,    119,    119,    119,    119,    119,    119,    119,    119,
                183,    183,    183,    183,    183,    183,    183,    183,    183,    183,    /* 0x3f50 */
                183,    183,    183,    183,    183,    183,    183,    183,    183,    183,
                183,    183,    183,    183,    183,    183,    183,    183,    183,    183,
                183,    183,    183,    183,    183,    183,    183,    183,    183,    183,
                255,    255,    255,    255,    255,    255,    255,    255,                    /* screenhole */
                63,     63,     63,     63,     63,     63,     63,     63,     63,     63,     /* 0x3f80 */
                63,     63,     63,     63,     63,     63,     63,     63,     63,     63,
                63,     63,     63,     63,     63,     63,     63,     63,     63,     63,
                63,     63,     63,     63,     63,     63,     63,     63,     63,     63,
                127,    127,    127,    127,    127,    127,    127,    127,    127,    127,    /* 0x3fa8 */
                127,    127,    127,    127,    127,    127,    127,    127,    127,    127,
                127,    127,    127,    127,    127,    127,    127,    127,    127,    127,
                127,    127,    127,    127,    127,    127,    127,    127,    127,    127,
                191,    191,    191,    191,    191,    191,    191,    191,    191,    191,    /* 0x3fd0 */
                191,    191,    191,    191,    191,    191,    191,    191,    191,    191,
                191,    191,    191,    191,    191,    191,    191,    191,    191,    191,
                191,    191,    191,    191,    191,    191,    191,    191,    191,    191,
                255,    255,    255,    255,    255,    255,    255,    255             /* screenhole */
              };


static unsigned int hresaddr[192] = {
                0x0000, 0x0400, 0x0800, 0x0c00, 0x1000, 0x1400, 0x1800, 0x1c00,
                0x0080, 0x0480, 0x0880, 0x0c80, 0x1080, 0x1480, 0x1880, 0x1c80,
                0x0100, 0x0500, 0x0900, 0x0d00, 0x1100, 0x1500, 0x1900, 0x1d00,
                0x0180, 0x0580, 0x0980, 0x0d80, 0x1180, 0x1580, 0x1980, 0x1d80,
                0x0200, 0x0600, 0x0a00, 0x0e00, 0x1200, 0x1600, 0x1a00, 0x1e00,
                0x0280, 0x0680, 0x0a80, 0x0e80, 0x1280, 0x1680, 0x1a80, 0x1e80,
                0x0300, 0x0700, 0x0b00, 0x0f00, 0x1300, 0x1700, 0x1b00, 0x1f00,
                0x0380, 0x0780, 0x0b80, 0x0f80, 0x1380, 0x1780, 0x1b80, 0x1f80,
                0x0028, 0x0428, 0x0828, 0x0c28, 0x1028, 0x1428, 0x1828, 0x1c28,
                0x00a8, 0x04a8, 0x08a8, 0x0ca8, 0x10a8, 0x14a8, 0x18a8, 0x1ca8,
                0x0128, 0x0528, 0x0928, 0x0d28, 0x1128, 0x1528, 0x1928, 0x1d28,
                0x01a8, 0x05a8, 0x09a8, 0x0da8, 0x11a8, 0x15a8, 0x19a8, 0x1da8,
                0x0228, 0x0628, 0x0a28, 0x0e28, 0x1228, 0x1628, 0x1a28, 0x1e28,
                0x02a8, 0x06a8, 0x0aa8, 0x0ea8, 0x12a8, 0x16a8, 0x1aa8, 0x1ea8,
                0x0328, 0x0728, 0x0b28, 0x0f28, 0x1328, 0x1728, 0x1b28, 0x1f28,
                0x03a8, 0x07a8, 0x0ba8, 0x0fa8, 0x13a8, 0x17a8, 0x1ba8, 0x1fa8,
                0x0050, 0x0450, 0x0850, 0x0c50, 0x1050, 0x1450, 0x1850, 0x1c50,
                0x00d0, 0x04d0, 0x08d0, 0x0cd0, 0x10d0, 0x14d0, 0x18d0, 0x1cd0,
                0x0150, 0x0550, 0x0950, 0x0d50, 0x1150, 0x1550, 0x1950, 0x1d50,
                0x01d0, 0x05d0, 0x09d0, 0x0dd0, 0x11d0, 0x15d0, 0x19d0, 0x1dd0,
                0x0250, 0x0650, 0x0a50, 0x0e50, 0x1250, 0x1650, 0x1a50, 0x1e50,
                0x02d0, 0x06d0, 0x0ad0, 0x0ed0, 0x12d0, 0x16d0, 0x1ad0, 0x1ed0,
                0x0350, 0x0750, 0x0b50, 0x0f50, 0x1350, 0x1750, 0x1b50, 0x1f50,
                0x03d0, 0x07d0, 0x0bd0, 0x0fd0, 0x13d0, 0x17d0, 0x1bd0, 0x1fd0
              };


	static unsigned char textcol[16] = {
		COL_LGR0,
		COL_LGR2,
		COL_LGR4,
		COL_LGR6,
		COL_LGR8,
		COL_LGRA,
		COL_LGRC,
		COL_LGRE,
		COL_LGR1,
		COL_LGR3,
		COL_LGR5,
		COL_LGR7,
		COL_LGR9,
		COL_LGRB,
		COL_LGRD,
		COL_LGRF
	      };

	static unsigned char textcol2[16] = {
		COL_LGR0,
		COL_LGR4,
		COL_LGR8,
		COL_LGRC,
		COL_LGR1,
		COL_LGR5,
		COL_LGR9,
		COL_LGRD,
		COL_LGR2,
		COL_LGR6,
		COL_LGRA,
		COL_LGRE,
		COL_LGR3,
		COL_LGR7,
		COL_LGRB,
		COL_LGRF
	      };


	static unsigned char textcol3[16] = {
		COL_LGR0,
		COL_LGR8,
		COL_LGR1,
		COL_LGR9,
		COL_LGR2,
		COL_LGRA,
		COL_LGR3,
		COL_LGRB,
		COL_LGR4,
		COL_LGRC,
		COL_LGR5,
		COL_LGRD,
		COL_LGR6,
		COL_LGRE,
		COL_LGR7,
		COL_LGRF
	      };

	static unsigned char textcol4[16] = {
		COL_LGR0,
		COL_LGR1,
		COL_LGR2,
		COL_LGR3,
		COL_LGR4,
		COL_LGR5,
		COL_LGR6,
		COL_LGR7,
		COL_LGR8,
		COL_LGR9,
		COL_LGRA,
		COL_LGRB,
		COL_LGRC,
		COL_LGRD,
		COL_LGRE,
		COL_LGRF
	      };

	static unsigned char hrescol[16] = {
		COL_LGR0,
		COL_LGR2,
		COL_LGR4,
		COL_LGR6,
		COL_LGR8,
		COL_LGRA,
		COL_LGRC,
		COL_LGRE,
		COL_LGR1,
		COL_LGR3,
		COL_LGR5,
		COL_LGR7,
		COL_LGR9,
		COL_LGRB,
		COL_LGRD,
		COL_LGRF
	      };

	static unsigned char hrescol2[16] = {
		COL_LGR0,
		COL_LGR4,
		COL_LGR8,
		COL_LGRC,
		COL_LGR1,
		COL_LGR5,
		COL_LGR9,
		COL_LGRD,
		COL_LGR2,
		COL_LGR6,
		COL_LGRA,
		COL_LGRE,
		COL_LGR3,
		COL_LGR7,
		COL_LGRB,
		COL_LGRF
	      };


	static unsigned char hrescol3[16] = {
		COL_LGR0,
		COL_LGR8,
		COL_LGR1,
		COL_LGR9,
		COL_LGR2,
		COL_LGRA,
		COL_LGR3,
		COL_LGRB,
		COL_LGR4,
		COL_LGRC,
		COL_LGR5,
		COL_LGRD,
		COL_LGR6,
		COL_LGRE,
		COL_LGR7,
		COL_LGRF
	      };

	static unsigned char hrescol4[16] = {
		COL_LGR0,
		COL_LGR1,
		COL_LGR2,
		COL_LGR3,
		COL_LGR4,
		COL_LGR5,
		COL_LGR6,
		COL_LGR7,
		COL_LGR8,
		COL_LGR9,
		COL_LGRA,
		COL_LGRB,
		COL_LGRC,
		COL_LGRD,
		COL_LGRE,
		COL_LGRF
	      };

	static unsigned char dhrscol[16] = {
		COL_LGR0,
		COL_LGR2,
		COL_LGR4,
		COL_LGR6,
		COL_LGR8,
		COL_LGRA,
		COL_LGRC,
		COL_LGRE,
		COL_LGR1,
		COL_LGR3,
		COL_LGR5,
		COL_LGR7,
		COL_LGR9,
		COL_LGRB,
		COL_LGRD,
		COL_LGRF
	      };

	static unsigned char dhrscol2[16] = {
		COL_LGR0,
		COL_LGR4,
		COL_LGR8,
		COL_LGRC,
		COL_LGR1,
		COL_LGR5,
		COL_LGR9,
		COL_LGRD,
		COL_LGR2,
		COL_LGR6,
		COL_LGRA,
		COL_LGRE,
		COL_LGR3,
		COL_LGR7,
		COL_LGRB,
		COL_LGRF
	      };


	static unsigned char dhrscol3[16] = {
		COL_LGR0,
		COL_LGR8,
		COL_LGR1,
		COL_LGR9,
		COL_LGR2,
		COL_LGRA,
		COL_LGR3,
		COL_LGRB,
		COL_LGR4,
		COL_LGRC,
		COL_LGR5,
		COL_LGRD,
		COL_LGR6,
		COL_LGRE,
		COL_LGR7,
		COL_LGRF
	      };

	static unsigned char dhrscol4[16] = {
		COL_LGR0,
		COL_LGR1,
		COL_LGR2,
		COL_LGR3,
		COL_LGR4,
		COL_LGR5,
		COL_LGR6,
		COL_LGR7,
		COL_LGR8,
		COL_LGR9,
		COL_LGRA,
		COL_LGRB,
		COL_LGRC,
		COL_LGRD,
		COL_LGRE,
		COL_LGRF
	      };



	static unsigned char dhrscol1rgb[16] = {
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR2,
		COL_LGR6,
		COL_LGRA,
		COL_LGRE,
		COL_LGR3,
		COL_LGR7,
		COL_LGRB,
		COL_LGRF
	      };


	static unsigned char dhrscol2rgb[16] = {
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR4,
		COL_LGRC,
		COL_LGR5,
		COL_LGRD,
		COL_LGR6,
		COL_LGRE,
		COL_LGR7,
		COL_LGRF
	      };

	static unsigned char dhrscol3rgb[16] = {
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR8,
		COL_LGR9,
		COL_LGRA,
		COL_LGRB,
		COL_LGRC,
		COL_LGRD,
		COL_LGRE,
		COL_LGRF
	      };

	static unsigned char dhrscol4rgb[16] = {
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR0,
		COL_LGR1,
		COL_LGR3,
		COL_LGR5,
		COL_LGR7,
		COL_LGR9,
		COL_LGRB,
		COL_LGRD,
		COL_LGRF
	      };

/*-------------------------------------*/


      void virtsetmode() {

	if (virtgraphic) {				/* set the display modes for both parts of the screen */
	  if (virthgr) {
	    if (virt80col) {
	      if (virtdhres) {
	        if (virtmonochrome) {
	          virtmodetop = MODEDHRSMONO;
	        }
	        else {
	          if (virtdhrsmode == 0) {
	            virtmodetop = MODEDHRS;
	          }
	          else {
	            if (virtdhrsmode == 1) {
	              virtmodetop = MODEDHRS140;
	            }
	            else {
	              virtmodetop = MODEDHRSRGB;
	            }
	          }
	        }
	        if (virtmixed) {
	          if ((virtmonochrome) || (virttextmode == 0)) {
	            virtmodedown = MODETEXT80;
	          }
	          else {
	            virtmodedown = MODETEXT80BLUR;
	          }
	          virtsplit = 160;
	        }
	        else {
	          virtmodedown = virtmodetop;
	          virtsplit = 192;
	        }
	      }
	      else { /* if (virtdhres) */
	        if (virtmonochrome) {
	          virtmodetop = MODEHRESMONO;
	        }
	        else {
	          if (virthresmode == 0) {
	            virtmodetop = MODEHRES;
	          }
	          else {
	            if (virthresmode == 1) {
	              virtmodetop = MODEHRESRGB;
	            }
	            else {
	              virtmodetop = MODEHRESBLUR;
	            }
	          }
	        }
	        if (virtmixed) {
	          if ((virtmonochrome) || (virttextmode == 0)) {
	            virtmodedown = MODETEXT80;
	          }
	          else {
	            virtmodedown = MODETEXT80BLUR;
	          }
	          virtsplit = 160;
	        }
	        else {
	          virtmodedown = virtmodetop;
	          virtsplit = 192;
	        }
	      }
	    }
	    else { /* if (virt80col */
	      if (virtmonochrome) {
	        virtmodetop = MODEHRESMONO;
	      }
	      else {
	        if (virtdhres) {
	          if (virthresmode == 0) {
	            virtmodetop = MODEHRESOFF;
	          }
	          else {
	            if (virthresmode == 1) {
	              virtmodetop = MODEHRESRGBOFF;
	            }
	            else {
	              virtmodetop = MODEHRESBLUROFF;
	            }
	          }
	        }
	        else { /* if (virtdhres) */
	          if (virthresmode == 0) {
	            virtmodetop = MODEHRES;
	          }
	          else {
	            if (virthresmode == 1) {
	              virtmodetop = MODEHRESRGB;
	            }
	            else {
	              virtmodetop = MODEHRESBLUR;
	            }
	          }
	        }
	      }
	      if (virtmixed) {
	        if ((virtmonochrome) || (virttextmode == 0)) { 
	          virtmodedown = MODETEXT40;
	        }
	        else {
	          virtmodedown = MODETEXT40BLUR;
	        }
	        virtsplit = 160;
	      }
	      else {
	        virtmodedown = virtmodetop;
	        virtsplit = 192;
	      }
	    }
	  }
	  else { /* if (virthgr) */
	    virtsplit = 0;
	    if (virt80col) {
	      if (virtdhres) {
	        if (virtmonochrome) {
	          virtmodetop = MODEDLRSMONO;
	        }
	        else {
	          virtmodetop = MODEDLRS;
	        }
	        if (virtmixed) {
	          if ((virtmonochrome) || (virttextmode == 0)) {
	            virtmodedown = MODETEXT80;
	          }
	          else {
	            virtmodedown = MODETEXT80BLUR;
	          }
	        }
	        else {
	          virtmodedown = virtmodetop;
	        }
	      }
	      else {
	        if (virtmonochrome) {
	          virtmodetop = MODELRESMONO;
	        }
	        else {
	          virtmodetop = MODELRES;
	        }
	        if (virtmixed) {
	          if ((virtmonochrome) || (virttextmode == 0)) {
	            virtmodedown = MODETEXT80;
	          }
	          else {
	            virtmodedown = MODETEXT80BLUR;
	          }
	        }
	        else {
	          virtmodedown = virtmodetop;
	        }
	      }
	    }
	    else {
	      if (virtmonochrome) {
	        virtmodetop = MODELRESMONO;
	      }
	      else {
	        virtmodetop = MODELRES;
	      }
	      if (virtmixed) {
	        if ((virtmonochrome) || (virttextmode == 0)) { 
	          virtmodedown = MODETEXT40;
	        }
	        else {
	          virtmodedown = MODETEXT40BLUR;
	        }
	      }
	      else {
	        virtmodedown = virtmodetop;
	      }
	    }
	  }
	}
	else {
	  virtsplit = 0;
	  if (virt80col) {
	    virtmodetop = MODETEXT80;
	    virtmodedown = virtmodetop;
	  }
	  else {
	    virtmodetop = MODETEXT40;
	    virtmodedown = virtmodetop;
	  }
	}

	if ((memstore80) || !(virtpage2)) {			/* set the visible page */
	  virtcachemode = virtcachemode & ~cachepage;
	  virttextpage = 0x400;
	  virthrespage = 0x2000;
	}
	else {
	  virtcachemode = virtcachemode | cachepage;
	  virttextpage = 0x800;
	  virthrespage = 0x4000;
	}

      } /* virtsetmode */


/*--------------------------------------*/


      void virtcacheinit() {

        memset(virtcache, 0, sizeof(virtcache));

      } /* virtcacheinit() */


/*--------------------------------------*/


      void virtsetcharset(Charset charset) {
	register unsigned int i;
	register unsigned int j;
	register unsigned char c;
	register unsigned char c2;
	static unsigned char xchr[11][12]={
	  {'#', '$', '@', '[', '\\',']', '^', '`', '{', '|', '}', '~'}, /* USA       */
	  {'#', '$', 133, 248, 135, 232, '^', '`', 130, 151, 138, 249}, /* France    */
	  {'#', '$', '@', 142, 153, 154, '^', '`', 132, 148, 129, 225}, /* Germany   */
	  {156, '$', '@', '[', '\\',']', '^', '`', '{', '|', '}', '~'}, /* UK        */
	  {'#', '$', '@', 146, 242, 143, '^', '`', 145, 243, 134, '~'}, /* Denmark 1 */
	  {'#', 241, 144, 142, 153, 143, 154, 130, 132, 148, 134, 129}, /* Sweden    */
	  {'#', '$', '@', 248, '\\',130, '^', 151, 133, 149, 138, 141}, /* Italy     */
	  {158, '$', '@', 173, 165, 168, '^', '`', 249, 164, '}', '~'}, /* Spain     */
	  {'#', '$', '@', '[', 157, ']', '^', '`', '{', '|', '}', '~'}, /* Japan     */
	  {'#', 241, 144, 146, 242, 143, 154, 130, 145, 243, 134, 129}, /* Norway    */
	  {'#', '$', 144, 146, 242, 143, 154, 130, 145, 243, 134, 129}, /* Denmark 2 */
	};
	static unsigned char mousechr[32]={
	  0x02, 0x01, 0xB2, 0x0F, 0xFB, 0x08, 0xA9, 0xAA,
	  0x1B, 0x0E, 0x19, 0x18, 0x15, 0x14, 0x16, 0x11,
	  0x10, 0x1F, 0x1E, 0xC4, 0x0D, 0x1A, 0xB0, 0xB1,
	  0xF4, 0xF5, 0x13, 0x04, 0xF0, 0x0A, 0x1C, 0x0C
	};


	switch (charset) {
	  case USA :
	    virtcharset = USA;
	    setmessage("Charset set to 'USA'");
	    break;
	  case France :
	    virtcharset = France;
	    setmessage("Charset set to 'France'");
	    break;
	  case Germany :
	    virtcharset = Germany;
	    setmessage("Charset set to 'Germany'");
	    break;
	  case UK :
	    virtcharset = UK;
	    setmessage("Charset set to 'United Kingdom'");
	    break;
	  case Denmark1 :
	    virtcharset = Denmark1;
	    setmessage("Charset set to 'Denmark1'");
	    break;
	  case Sweden :
	    virtcharset = Sweden;
	    setmessage("Charset set to 'Sweden'");
	    break;
	  case Italy :
	    virtcharset = Italy;
	    setmessage("Charset set to 'Italy'");
	    break;
	  case Spain :
	    virtcharset = Spain;
	    setmessage("Charset set to 'Spain'");
	    break;
	  case Japan :
	    virtcharset = Japan;
	    setmessage("Charset set to 'Japan'");
	    break;
	  case Norway :
	    virtcharset = Norway;
	    setmessage("Charset set to 'Norway'");
	    break;
	  case Denmark2 :
	    virtcharset = Denmark2;
	    setmessage("Charset set to 'Denmark2'");
	    break;
	  default :
	    return;
	} /* switch */

	for (i=0x100; i<0x400; i++) {
	  virtfont[i+0x400]	= fontoriginal[i];
	} /* for i */

	for (i=0; i<0x20; i++) {
	  for (j=0; j<8; j++) {
	    virtfont[(i << 3) + 0x400 + j] = fontoriginal[(mousechr[i] << 3) + j]; 
	  } /* for j */
	} /* for i */

	for (i=0; i<12; i++) {		/* replace 8 characters with country specific chars */
	  c = xchr[charset][i];
	  for (j=0; j<8; j++) {
	    virtfont[ (xchr[0][i] << 3) + 0x400 + j] = fontoriginal[(c << 3) + j];
	  } /* for j */
	} /* for i */

	for (i=0x400; i<0x800; i++) {
	  c = virtfont[i];
	  c2 = 0;
	  if (c & 0x80) { c2 = c2 | 0x01; }
	  if (c & 0x40) { c2 = c2 | 0x02; }
	  if (c & 0x20) { c2 = c2 | 0x04; }
	  if (c & 0x10) { c2 = c2 | 0x08; }
	  if (c & 0x08) { c2 = c2 | 0x10; }
	  if (c & 0x04) { c2 = c2 | 0x20; }
	  if (c & 0x02) { c2 = c2 | 0x40; }
	  virtfont[i] = c2;
	} /* for i */

	for (i=0; i<0x100; i++) {
	  c = virtfont [i+0x400];		/* mousetext */
	  virtfontalt  [i+0x200]= c;		/* on enhanced Apple//e only */

	  c = virtfont [i+0x500];		/* numbers */
	  virtfont     [i+0x500]= c;
	  virtfontflash[i+0x300]= c;
	  virtfontflash[i+0x500]= c;
	  virtfontalt  [i+0x500]= c;
	  c = ~c;
	  virtfont     [i+0x100]= c;
	  virtfont     [i+0x300]= c;
	  virtfontflash[i+0x100]= c;
	  virtfontalt  [i+0x100]= c;

	  c = virtfont [i+0x600];		/* capital chars */
	  virtfont     [i+0x400]= c;
	  virtfont     [i+0x600]= c;
	  virtfontflash[i+0x200]= c;
	  virtfontflash[i+0x400]= c;
	  virtfontflash[i+0x600]= c;
	  virtfontalt  [i+0x400]= c;
	  virtfontalt  [i+0x600]= c;
	  c = ~c;
	  virtfont     [i+0x000]= c;
	  virtfont     [i+0x200]= c;
	  virtfontflash[i+0x000]= c;
	  virtfontalt  [i+0x000]= c;
/*	  virtfontalt  [i+0x200]= c;		on unenhanced AppleIIe only */

	  c = virtfont [i+0x700];		/* small chars */
	  if (appletype == APPLEII) {
	    c2 = virtfont[i+0x600];
	    virtfont     [i+0x700]= c2;
	    virtfontflash[i+0x700]= c2;
	  }
	  else {
	    virtfont     [i+0x700]= c;
	    virtfontflash[i+0x700]= c;
	  }
	  virtfontalt  [i+0x700]= c;
	  c = ~c;
	  virtfontalt  [i+0x300]= c;

	} /* for i */

	virtfontptr	 = virtfont;
	virtfontflashptr = virtfont;
      } /* virtsetcharset */


/*-------------------------------------*/


      Charset virtgetcharset() {

	return virtcharset;

      } /* virtgetcharset */


/*-------------------------------------*/


      void virtsettextmode(unsigned char mode) {

	if (mode == 0) {
	  virttextmode = 0;
	  setmessage("Set text mode to 'Standard'");
	  virtsetmode();
	  virtcacheinit();
	}
	else {
	  if (mode == 1) {
	    virttextmode = 1;
	    setmessage("Set text mode to 'Blurry'");
	    virtsetmode();
	    virtcacheinit();
	  }
	}

      } /* virtsettextmode */


/*-------------------------------------*/


      unsigned char virtgettextmode() {

	return virttextmode;

      } /* virtgettextmode */


/*-------------------------------------*/


      void virtsethresmode(unsigned char mode) {

	switch (mode) {
	  case 0 :	/* Set standard */
	    virthresmode = 0;
		/* even columns */
	    colourtabe[0] = COL_HGR0;	/* 000 */
	    colourtabe[1] = COL_HGR0;	/* 100 */
	    colourtabe[2] = COL_HGR2;	/* 010 */
	    colourtabe[3] = COL_HGR3;	/* 110 */
	    colourtabe[4] = COL_HGR0;	/* 001 */
	    colourtabe[5] = COL_HGR1;	/* 101 */
	    colourtabe[6] = COL_HGR3;	/* 011 */
	    colourtabe[7] = COL_HGR3;	/* 111 */
		/* odd columns */
	    colourtabo[0] = COL_HGR0;	/* 000 */
	    colourtabo[1] = COL_HGR0;	/* 100 */
	    colourtabo[2] = COL_HGR1;	/* 010 */
	    colourtabo[3] = COL_HGR3;	/* 110 */
	    colourtabo[4] = COL_HGR0;	/* 001 */
	    colourtabo[5] = COL_HGR2;	/* 101 */
	    colourtabo[6] = COL_HGR3;	/* 011 */
	    colourtabo[7] = COL_HGR3;	/* 111 */
		/* even columns */
	    colourtabe2[0] = COL_HGR4;	/* 000 */
	    colourtabe2[1] = COL_HGR4;	/* 100 */
	    colourtabe2[2] = COL_HGR6;	/* 010 */
	    colourtabe2[3] = COL_HGR7;	/* 110 */
	    colourtabe2[4] = COL_HGR4;	/* 001 */
	    colourtabe2[5] = COL_HGR5;	/* 101 */
	    colourtabe2[6] = COL_HGR7;	/* 011 */
	    colourtabe2[7] = COL_HGR7;	/* 111 */
		/* odd columns */
	    colourtabo2[0] = COL_HGR4;	/* 000 */
	    colourtabo2[1] = COL_HGR4;	/* 100 */
	    colourtabo2[2] = COL_HGR5;	/* 010 */
	    colourtabo2[3] = COL_HGR7;	/* 110 */
	    colourtabo2[4] = COL_HGR4;	/* 001 */
	    colourtabo2[5] = COL_HGR6;	/* 101 */
	    colourtabo2[6] = COL_HGR7;	/* 011 */
	    colourtabo2[7] = COL_HGR7;	/* 111 */
	    setmessage("Set hi-res mode to 'Composite'");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  case 1 :	/* Set RGB */
	    virthresmode = 1;
		/* even columns */
	    colourtabe[0] = COL_HGR0;	/* 000 */
	    colourtabe[1] = COL_HGR0;	/* 100 */
	    colourtabe[2] = COL_HGR2;	/* 010 */
	    colourtabe[3] = COL_HGR3;	/* 110 */
	    colourtabe[4] = COL_HGR0;	/* 001 */
	    colourtabe[5] = COL_HGR0;	/* 101 */
	    colourtabe[6] = COL_HGR2;	/* 011 */
	    colourtabe[7] = COL_HGR3;	/* 111 */
		/* odd columns */
	    colourtabo[0] = COL_HGR0;	/* 000 */
	    colourtabo[1] = COL_HGR0;	/* 100 */
	    colourtabo[2] = COL_HGR1;	/* 010 */
	    colourtabo[3] = COL_HGR3;	/* 110 */
	    colourtabo[4] = COL_HGR0;	/* 001 */
	    colourtabo[5] = COL_HGR0;	/* 101 */
	    colourtabo[6] = COL_HGR1;	/* 011 */
	    colourtabo[7] = COL_HGR3;	/* 111 */
		/* even columns */
	    colourtabe2[0] = COL_HGR4;	/* 000 */
	    colourtabe2[1] = COL_HGR4;	/* 100 */
	    colourtabe2[2] = COL_HGR6;	/* 010 */
	    colourtabe2[3] = COL_HGR7;	/* 110 */
	    colourtabe2[4] = COL_HGR4;	/* 001 */
	    colourtabe2[5] = COL_HGR4;	/* 101 */
	    colourtabe2[6] = COL_HGR6;	/* 011 */
	    colourtabe2[7] = COL_HGR7;	/* 111 */
		/* odd columns */
	    colourtabo2[0] = COL_HGR4;	/* 000 */
	    colourtabo2[1] = COL_HGR4;	/* 100 */
	    colourtabo2[2] = COL_HGR5;	/* 010 */
	    colourtabo2[3] = COL_HGR7;	/* 110 */
	    colourtabo2[4] = COL_HGR4;	/* 001 */
	    colourtabo2[5] = COL_HGR4;	/* 101 */
	    colourtabo2[6] = COL_HGR5;	/* 011 */
	    colourtabo2[7] = COL_HGR7;	/* 111 */
	    setmessage("Set hi-res mode to 'RGB'");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  case 2 :
	    virthresmode = 2;
	    setmessage("Set hi-res mode to 'Blurry'");
	    virtsetmode();
	    virtcacheinit();
	    break;
	} /* switch */

      } /* virtsethresmode */


/*-------------------------------------*/


      unsigned char virtgethresmode() {

	return virthresmode;

      } /* virtgethresmode */


/*-------------------------------------*/


      void virtsetdhrsmode(unsigned char mode) {

	if (mode == 0) {
	  virtdhrsmode = 0;
	  setmessage("Set Double Hi-Res resolution to 560 * 192 pixels");
	  virtsetmode();
	  virtcacheinit();
	}
	else {
	  if (mode == 1) {
	    virtdhrsmode = 1;
	    setmessage("Set Double Hi-Res resolution to 140 * 192 pixels");
	    virtsetmode();
	    virtcacheinit();
	  }
	  else {
	    if (mode == 2) {
	      virtdhrsmode = 2;
	      setmessage("Set Double Hi-Res resolution to 'RGB'");
	      virtsetmode();
	      virtcacheinit();
	    }
	  }
	}

      } /* virtsetdhrsmode */


/*-------------------------------------*/


      unsigned char virtgetdhrsmode() {

	return virtdhrsmode;

      } /* virtgetdhrsmode */


/*-------------------------------------*/


      void virtsetmonochrome(unsigned char mode) {

	switch (mode) {
	  case 0 :	/* colour */
	    virtmonochrome = 0;
	    virtmonocol0 = COL_TXT_WHT0;
	    virtmonocol1 = COL_TXT_WHT1;
	    setmessage("Set color mode");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  case 1 :	/* monochrome white */
	    virtmonochrome = 1;
	    virtmonocol0 = COL_TXT_WHT0;
	    virtmonocol1 = COL_TXT_WHT1;
	    setmessage("Set monochrome mode white");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  case 2 :	/* monochrome green */
	    virtmonochrome = 2;
	    virtmonocol0 = COL_TXT_GRN0;
	    virtmonocol1 = COL_TXT_GRN1;
	    setmessage("Set monochrome mode green");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  case 3 :	/* monochrome amber */
	    virtmonochrome = 3;
	    virtmonocol0 = COL_TXT_AMB0;
	    virtmonocol1 = COL_TXT_AMB1;
	    setmessage("Set monochrome mode amber");
	    virtsetmode();
	    virtcacheinit();
	    break;
	  } /* switch */

      } /* virtsetmonochrome */


/*-------------------------------------*/


      unsigned char virtgetmonochrome() {

	return virtmonochrome;

      } /* virtgetmonochrome */


/*-------------------------------------*/


      void virtpaletteset(unsigned int index, unsigned int r,
			  unsigned int g,     unsigned int b) {

	imagesetpcolor(window, index, (r<<16)|(g<<8)|b);
	imagesetpcolor(window, index + 0x20, (((r<<16)|(g<<8)|b) & 0xfefefe) >> 1);

      } /* virtpaletteset */


/*-------------------------------------*/


      void virtsetpalette(unsigned char mode) {

	switch (mode) {
	  case 0x00:			/* AppleIIgs */
	    virtpalette = 0;
	    virtpaletteset(COL_LGR0,   0,  0,  0);
	    virtpaletteset(COL_LGR1, 221,  0, 51);
	    virtpaletteset(COL_LGR2,   0,  0,153);
	    virtpaletteset(COL_LGR3, 221, 34,221);
	    virtpaletteset(COL_LGR4,   0,119, 34);
	    virtpaletteset(COL_LGR5,  85, 85, 85);
	    virtpaletteset(COL_LGR6,  34, 34,255);
	    virtpaletteset(COL_LGR7, 102,170,255);
	    virtpaletteset(COL_LGR8, 136, 85,  0);
	    virtpaletteset(COL_LGR9, 255,102,  0);
	    virtpaletteset(COL_LGRA, 170,170,170);
	    virtpaletteset(COL_LGRB, 255,153,136);
	    virtpaletteset(COL_LGRC,  17,221,  0);
	    virtpaletteset(COL_LGRD, 255,255,  0);
	    virtpaletteset(COL_LGRE,  68,255,153);
	    virtpaletteset(COL_LGRF, 255,255,255);
	    virtpaletteset(COL_HGR0,   0,  0,  0);
	    virtpaletteset(COL_HGR1,  17,221,  0);
	    virtpaletteset(COL_HGR2, 221, 34,221);
	    virtpaletteset(COL_HGR3, 255,255,255);
	    virtpaletteset(COL_HGR4,   0,  0,  0);
	    virtpaletteset(COL_HGR5, 255,102,  0);
	    virtpaletteset(COL_HGR6,  34, 34,255);
	    virtpaletteset(COL_HGR7, 255,255,255);
	    setmessage("Set palette type 'AppleIIgs'");
	    virtcopy = 1;
	    break;
	  case 0x01:			/* Standard */
	    virtpalette = 1;
	    virtpaletteset(COL_LGR0,   0,  0,  0);
	    virtpaletteset(COL_LGR1, 227, 30, 96);
	    virtpaletteset(COL_LGR2,  96, 78,189);
	    virtpaletteset(COL_LGR3, 255, 68,253);
	    virtpaletteset(COL_LGR4,   0,163, 96);
	    virtpaletteset(COL_LGR5, 156,156,156);
	    virtpaletteset(COL_LGR6,  20,207,253);
	    virtpaletteset(COL_LGR7, 208,195,255);
	    virtpaletteset(COL_LGR8,  96,114,  3);
	    virtpaletteset(COL_LGR9, 255,106, 60);
	    virtpaletteset(COL_LGRA, 156,156,156);
	    virtpaletteset(COL_LGRB, 255,160,208);
	    virtpaletteset(COL_LGRC,  20,245, 60);
	    virtpaletteset(COL_LGRD, 208,221,141);
	    virtpaletteset(COL_LGRE, 114,255,208);
	    virtpaletteset(COL_LGRF, 255,255,255);
	    virtpaletteset(COL_HGR0,   0,  0,  0);
	    virtpaletteset(COL_HGR1,  20,245, 60);
	    virtpaletteset(COL_HGR2, 255, 68,253);
	    virtpaletteset(COL_HGR3, 255,255,255);
	    virtpaletteset(COL_HGR4,   0,  0,  0);
	    virtpaletteset(COL_HGR5, 255,106, 60);
	    virtpaletteset(COL_HGR6,  20,207,253);
	    virtpaletteset(COL_HGR7, 255,255,255);
	    setmessage("Set palette type 'Standard'");
	    virtcopy = 1;
	    break;
	  case 0x02:			/* ApplePC */
	    virtpalette = 2;
	    virtpaletteset(COL_LGR0,   0,  0,  0);
	    virtpaletteset(COL_LGR1, 124, 32, 64);
	    virtpaletteset(COL_LGR2,  32, 48,124);
	    virtpaletteset(COL_LGR3, 188, 80,188);
	    virtpaletteset(COL_LGR4,   0, 92, 60);
	    virtpaletteset(COL_LGR5, 124,124,124);
	    virtpaletteset(COL_LGR6,  64,140,184);
	    virtpaletteset(COL_LGR7, 188,172,248);
	    virtpaletteset(COL_LGR8,  60, 76,  0);
	    virtpaletteset(COL_LGR9, 184,108, 64);
	    virtpaletteset(COL_LGRA, 124,124,124);
	    virtpaletteset(COL_LGRB, 248,156,188);
	    virtpaletteset(COL_LGRC,  60,168, 60);
	    virtpaletteset(COL_LGRD, 188,200,124);
	    virtpaletteset(COL_LGRE, 124,216,188);
	    virtpaletteset(COL_LGRF, 255,255,255);
	    virtpaletteset(COL_HGR0,   0,  0,  0);
	    virtpaletteset(COL_HGR1,  60,168, 60);
	    virtpaletteset(COL_HGR2, 188, 80,188);
	    virtpaletteset(COL_HGR3, 224,224,224);
	    virtpaletteset(COL_HGR4,   0,  0,  0);
	    virtpaletteset(COL_HGR5, 184,108, 64);
	    virtpaletteset(COL_HGR6,  64,140,184);
	    virtpaletteset(COL_HGR7, 224,224,224);
	    setmessage("Set palette type 'ApplePC'");
	    virtcopy = 1;
	    break;
          case 0x03:
            virtpalette = 3;
            virtpaletteset(COL_LGR0,   0,  0,  0);
            virtpaletteset(COL_LGR1,  96, 96, 96);
            virtpaletteset(COL_LGR2,  95, 95, 95);
            virtpaletteset(COL_LGR3, 144,144,144);
            virtpaletteset(COL_LGR4, 107,107,107);
            virtpaletteset(COL_LGR5, 156,156,156);
            virtpaletteset(COL_LGR6, 155,155,155);
            virtpaletteset(COL_LGR7, 205,205,205);
            virtpaletteset(COL_LGR8,  97, 97, 97);
            virtpaletteset(COL_LGR9, 146,146,146);
            virtpaletteset(COL_LGRA, 155,155,155);
            virtpaletteset(COL_LGRB, 194,194,194);
            virtpaletteset(COL_LGRC, 157,157,157);
            virtpaletteset(COL_LGRD, 208,208,208);
            virtpaletteset(COL_LGRE, 207,207,207);
            virtpaletteset(COL_LGRF, 255,255,255);
            virtpaletteset(COL_HGR0,   0,  0,  0);
            virtpaletteset(COL_HGR1, 157,157,157);
            virtpaletteset(COL_HGR2, 144,144,144);
            virtpaletteset(COL_HGR3, 255,255,255);
            virtpaletteset(COL_HGR4,   0,  0,  0);
            virtpaletteset(COL_HGR5, 146,146,146);
            virtpaletteset(COL_HGR6, 155,155,155);
            virtpaletteset(COL_HGR7, 255,255,255);
            setmessage("Set palette type 'Grayscale'");
            virtcopy = 1;
            break;
	} /* switch */

      } /* virtsetpalette */


/*-------------------------------------*/


      unsigned char virtgetpalette() {

	return virtpalette;

      } /* virtgetpalette */


/*-------------------------------------*/


      void virtpaletteinit() {

	virtsetpalette(1);	/* standard colours for LoRes and HiRes */

	virtpaletteset(COL_TXT_WHT0,0x00,0x00,0x00);
	virtpaletteset(COL_TXT_WHT1,0xff,0xff,0xff);
	virtpaletteset(COL_TXT_GRN0,0x00,0x00,0x00);
	virtpaletteset(COL_TXT_GRN1,0x00,0xaf,0x00);
	virtpaletteset(COL_TXT_AMB0,0x00,0x00,0x00);
	virtpaletteset(COL_TXT_AMB1,0xbf,0x5f,0x00);

      } /* virtpaletteinit */


/*-------------------------------------*/


      void virtreset () {

	virtgraphic	= 0;
	virtmixed	= 0;
	virtpage2	= 0;
	virthgr		= 0;
	virt80col	= 0;
	virtaltchar	= 0;
	virtiou		= 0x80;
	virtdhres	= 0;
	virtflashmode	= cachetext40;
	virtflashmodeb	= cachetext40b;
	virtcacheinit();
	virtsetmode();

      } /* virtreset */


/*-------------------------------------*/


      void virtinit() {
	register unsigned int i;
	register unsigned int byte;


	for (i=0; i<0x100; i++) {
	  byte = 0;
	  if (i&0x80) { byte = byte | 0xc000; }
	  if (i&0x40) { byte = byte | 0x3000; }
	  if (i&0x20) { byte = byte | 0x0c00; }
	  if (i&0x10) { byte = byte | 0x0300; }
	  if (i&0x08) { byte = byte | 0x00c0; }
	  if (i&0x04) { byte = byte | 0x0030; }
	  if (i&0x02) { byte = byte | 0x000c; }
	  if (i&0x01) { byte = byte | 0x0003; }
	  doublebits[i] = byte;
	} /* for i */

	virtsetcharset(USA);
	virtcacheinit();			/* clear cache */
	virtcachemode = 0;

	virtcopy      = 0;

	virtsetmonochrome(0);			/* set to colour mode */
	virtsettextmode(0);			/* set standard text mode */
	virtsethresmode(0);			/* set standard hi-res mode */
	virtsetdhrsmode(0);			/* set standard double hi-res mode */
	virtpaletteinit();			/* set standard palette */
	virtscanline = 0;
	virtdouble = 0;

	virtreset();				/* reset softswitches */
	virtsetmode();				/* set mode and page for rasterline */

      } /* virtinit */


/*-------------------------------------*/


      unsigned int virtstore(unsigned int window, FILE *file) {

	fwrite(&virtgraphic,	sizeof(virtgraphic),	1, file);
	fwrite(&virtmixed,	sizeof(virtmixed),	1, file);
	fwrite(&virtpage2,	sizeof(virtpage2),	1, file);
	fwrite(&virthgr,	sizeof(virthgr),	1, file);
	fwrite(&virt80col,	sizeof(virt80col),	1, file);
	fwrite(&virtaltchar,	sizeof(virtaltchar),	1, file);
	fwrite(&virtiou,	sizeof(virtiou),	1, file);
	fwrite(&virtdhres,	sizeof(virtdhres),	1, file);
	fwrite(&virtcharset,	sizeof(virtcharset),	1, file);
	fwrite(&virttextmode,	sizeof(virttextmode),	1, file);
	fwrite(&virthresmode,	sizeof(virthresmode),	1, file);
	fwrite(&virtdhrsmode,	sizeof(virtdhrsmode),	1, file);
	fwrite(&virtmonochrome,	sizeof(virtmonochrome),	1, file);
	fwrite(&virtpalette,	sizeof(virtpalette),	1, file);
	fwrite(&virtdouble,	sizeof(virtdouble),	1, file);
	fwrite(&virtscanline,	sizeof(virtscanline),	1, file);

	stringwrite(window, "Video stored.\r");

	return 0;

      } /* virtstore */


/*-------------------------------------*/


      unsigned int virtrestore(unsigned int window, FILE *file) {

	fread(&virtgraphic,	sizeof(virtgraphic),	1, file);
	fread(&virtmixed,	sizeof(virtmixed),	1, file);
	fread(&virtpage2,	sizeof(virtpage2),	1, file);
	fread(&virthgr,		sizeof(virthgr),	1, file);
	fread(&virt80col,	sizeof(virt80col),	1, file);
	fread(&virtaltchar,	sizeof(virtaltchar),	1, file);
	fread(&virtiou,		sizeof(virtiou),	1, file);
	fread(&virtdhres,	sizeof(virtdhres),	1, file);
	fread(&virtcharset,	sizeof(virtcharset),	1, file);
	virtsetcharset(virtcharset);
	fread(&virttextmode,	sizeof(virttextmode),	1, file);
	virtsettextmode(virttextmode);
	fread(&virthresmode,	sizeof(virthresmode),	1, file);
	virtsethresmode(virthresmode);
	fread(&virtdhrsmode,	sizeof(virtdhrsmode),	1, file);
	virtsetdhrsmode(virtdhrsmode);
	fread(&virtmonochrome,	sizeof(virtmonochrome),	1, file);
	virtsetmonochrome(virtmonochrome);
	fread(&virtpalette,	sizeof(virtpalette),	1, file);
	virtsetpalette(virtpalette);
	fread(&virtdouble,	sizeof(virtdouble),	1, file);
	fread(&virtscanline,	sizeof(virtscanline),	1, file);
	virtsetmode();
	virtcacheinit();

	stringwrite(window, "Video restored.\r");

	return 0;

      } /* virtrestore */


/*-------------------------------------*/


      unsigned int virtcalcline (unsigned char *virtptr, unsigned int rastline) {
	unsigned int addr, mode;
	register unsigned int val;
	unsigned char val8;
	unsigned char bit, bytes;

	if (rastline <= 159) {
	  val = virtmodetop;
	}
	else {
	  val = virtmodedown;
	}

	switch (val) {

	  case MODETEXT40 :		/* text 40 column */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr];
	    val = virtcachemode | virtflashmode;
	    if (virtcache[rastline] == val) { return 0; }
	    virtcache[rastline] = val;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val = *(virtfontptr+ ((memram[addr++] << 3) | (rastline & 7)) );
	      for ( bit = 1; bit < 0x80; bit=bit<<1) {
		val8 = (val & bit) ? virtmonocol1 : virtmonocol0;
		*virtptr++ = val8;
		*virtptr++ = val8;
	      } /* for bit */
	    }	/* for bytes */
	    return -1;

	  case MODETEXT40BLUR :		/* text 40 column blur */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr];
	    val = virtcachemode | virtflashmodeb;
	    if (virtcache[rastline] == val) { return 0; }
	    virtcache[rastline] = val;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val = (doublebits[*(virtfontptr+ ((memram[addr  ] << 3) | (rastline & 7)) )])
	          | (doublebits[*(virtfontptr+ ((memram[addr+1] << 3) | (rastline & 7)) )] << 14)
	          | (doublebits[*(virtfontptr+ ((memram[addr+2] << 3) | (rastline & 7)) )] << 28);
	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = textcol4[ val & 0xf ];
	      *virtptr++ = textcol [ (val >>  1) & 0xf ];
	      *virtptr++ = textcol2[ (val >>  2) & 0xf ];
	      *virtptr++ = textcol3[ (val >>  3) & 0xf ];
	      *virtptr++ = textcol4[ (val >>  4) & 0xf ];
	      *virtptr++ = textcol [ (val >>  5) & 0xf ];
	      *virtptr++ = textcol2[ (val >>  6) & 0xf ];
	      *virtptr++ = textcol3[ (val >>  7) & 0xf ];
	      *virtptr++ = textcol4[ (val >>  8) & 0xf ];
	      *virtptr++ = textcol [ (val >>  9) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 10) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 11) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 12) & 0xf ];
	      *virtptr++ = textcol [ (val >> 13) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 14) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 15) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 16) & 0xf ];
	      *virtptr++ = textcol [ (val >> 17) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 18) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 19) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 20) & 0xf ];
	      *virtptr++ = textcol [ (val >> 21) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 22) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 23) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 24) & 0xf ];
	      *virtptr++ = textcol [ (val >> 25) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 26) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 27) & 0xf ];
	      addr++;
	      addr++;
	    } /* for bytes */
	   return -1;

	  case MODETEXT80 :		/* text 80 column */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachetext80;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val8 = virtfontalt[(memram[addr+0x10000] << 3) | (rastline & 7)];
	      *virtptr++ = (val8 & 0x01) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x02) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x04) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x08) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x10) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x20) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x40) ? virtmonocol1 : virtmonocol0;
	      val8 = virtfontalt[(memram[addr++] << 3) | (rastline & 7)];
	      *virtptr++ = (val8 & 0x01) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x02) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x04) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x08) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x10) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x20) ? virtmonocol1 : virtmonocol0;
	      *virtptr++ = (val8 & 0x40) ? virtmonocol1 : virtmonocol0;
	    } /* for bytes */
	    return -1;

	  case MODETEXT80BLUR :		/* text 80 column blur */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachetext80;
	    if (virtcache[rastline] == val) { return 0; }
	    virtcache[rastline] = val;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val = (virtfontalt[(memram[addr+0x10000] << 3) | (rastline & 7)]      )
	          | (virtfontalt[(memram[addr        ] << 3) | (rastline & 7)] <<  7)
	          | (virtfontalt[(memram[addr+0x10001] << 3) | (rastline & 7)] << 14)
	          | (virtfontalt[(memram[addr+1      ] << 3) | (rastline & 7)] << 21)
	          | (virtfontalt[(memram[addr+0x10002] << 3) | (rastline & 7)] << 28);
	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = textcol4[ val & 0xf ];
	      *virtptr++ = textcol [ (val >>  1) & 0xf ];
	      *virtptr++ = textcol2[ (val >>  2) & 0xf ];
	      *virtptr++ = textcol3[ (val >>  3) & 0xf ];
	      *virtptr++ = textcol4[ (val >>  4) & 0xf ];
	      *virtptr++ = textcol [ (val >>  5) & 0xf ];
	      *virtptr++ = textcol2[ (val >>  6) & 0xf ];
	      *virtptr++ = textcol3[ (val >>  7) & 0xf ];
	      *virtptr++ = textcol4[ (val >>  8) & 0xf ];
	      *virtptr++ = textcol [ (val >>  9) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 10) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 11) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 12) & 0xf ];
	      *virtptr++ = textcol [ (val >> 13) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 14) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 15) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 16) & 0xf ];
	      *virtptr++ = textcol [ (val >> 17) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 18) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 19) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 20) & 0xf ];
	      *virtptr++ = textcol [ (val >> 21) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 22) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 23) & 0xf ];
	      *virtptr++ = textcol4[ (val >> 24) & 0xf ];
	      *virtptr++ = textcol [ (val >> 25) & 0xf ];
	      *virtptr++ = textcol2[ (val >> 26) & 0xf ];
	      *virtptr++ = textcol3[ (val >> 27) & 0xf ];
	      addr++;
	      addr++;
	    } /* for bytes */
	   return -1;

	  case MODELRESMONO :		/* lores 40 column mono */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachelres40;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    if ((rastline & 7) < 4) {
	      for (bytes = 20; bytes; bytes--) {
		val8 = memram[addr++] & 0xf;
		val = (val8 << 12) |(val8 << 8) |(val8 << 4) | val8;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
		val8 = memram[addr++] & 0xf;
		val = ((val8 << 12) |(val8 << 8) |(val8 << 4) | val8) >> 2;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
	      } /* for bytes */
	      return -1;
	    }
	    else {
	      for (bytes = 20; bytes; bytes--) {
		val8 = (memram[addr++] & 0xf0) >> 4;
		val = (val8 << 12) |(val8 << 8) |(val8 << 4) | val8;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
		val8 = (memram[addr++] & 0xf0) >> 4;
		val = ((val8 << 12) |(val8 << 8) |(val8 << 4) | val8) >> 2;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
	      } /* for bytes */
	      return -1;
	    }

	  case MODELRES :		/* lores 40 column */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachelres40;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    if ((rastline & 7) < 4) {
	      for (bytes = 40; bytes; bytes--) {
		val8 = (memram[addr++] & 0xf) + COL_LGR0;
		for (bit=14; bit; bit--) {
	          *virtptr++ = val8;
	        } /* for bit */
	      } /* for bytes */
	      return -1;
	    }
	    else {
	      for (bytes = 40; bytes; bytes--) {
		val8 = (memram[addr++] >> 4) + COL_LGR0;
		 for (bit=14; bit; bit--) {
		  *virtptr++ = val8;
		} /* for bit */
	      } /* for bytes */
	      return -1;
	    }

	  case MODEDLRSMONO :		/* lores 80 column mono */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachelres80;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    if ((rastline & 7) < 4) {
	      for (bytes = 20; bytes; bytes--) {
		val8 = memram[addr+0x10000] & 0xf;
		val = (val8 << 4) | val8;
		val8 = memram[addr++] & 0xf;
		val = (val8 << 12) | (val8 << 8) | val;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
		val8 = memram[addr+0x10000] & 0xf;
		val = (val8 << 4) | val8;
		val8 = memram[addr++] & 0xf;
		val = ((val8 << 12) | (val8 << 8) | val) >> 2;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
	      } /* for bytes */
	      return -1;
	    }
	    else {
	      for (bytes = 20; bytes; bytes--) {
		val8 = memram[addr+0x10000] & 0xf0;
		val = (val8 >> 4) | val8;
		val8 = memram[addr++] & 0xf0;
		val = (val8 << 8) | (val8 << 4) | val;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
		val8 = memram[addr+0x10000] & 0xf0;
		val = (val8 >> 4) | val8;
		val8 = memram[addr++] & 0xf0;
		val = (val8 << 8) | (val8 << 4) | val;
		for (mode=1; mode<0x4000; mode = mode << 1) {
		  *virtptr++ = (val & mode) ? virtmonocol1 : virtmonocol0;
		} /* for mode */
	      } /* for bytes */
	      return -1;
	    }

	  case MODEDLRS :		/* lores 80 column */
	    addr = textaddr[rastline] + virttextpage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachelres80;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    if ((rastline & 7) < 4) {
	      for (bytes = 40; bytes; bytes--) {
		val8 = (memram[addr+0x10000] & 0xf) + COL_LGR0;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = (memram[addr++] & 0xf) + COL_LGR0;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
	      } /* for bytes */
	      return -1;
	    }
	    else {
	      for (bytes = 40; bytes; bytes--) {
		val8 = (memram[addr+65536] >> 4) + COL_LGR0;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = (memram[addr++] >> 4) + COL_LGR0;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
		*virtptr++ = val8;
	      } /* for bytes */
	      return -1;
	    }

	  case MODEHRESMONO :		/* hires mono */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehres;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val = memram[addr++];
	      for ( bit = 1; bit<128; bit=bit<<1) {
		val8 = (val & bit) ? virtmonocol1 : virtmonocol0;
		*virtptr++ = val8;
		*virtptr++ = val8;
	      } /* for bit */
	    } /* for bytes */
	    return -1;

	  case MODEHRES :		/* hires standard */
	  case MODEHRESRGB :		/* hires RGB */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehres;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val = ((memram[addr-1] & 0x60) >> 5)
		  | ((memram[addr]   & 0x7f) << 2)
		  | ((memram[addr+1] & 0x3)  << 9);
	      if ((bytes & 1) == 0) {              /* even column */
		if (bytes == 40) {
		  val = val & 0x7fc;       /* drop surrounding bits at left border */
		}
		if (memram[addr] & 0x80) {
		  val8 = colourtabe2[(val >> 1) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 2) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 3) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 4) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 5) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 6) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 7) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		}
		else {
		  val8 = colourtabe[(val >> 1) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 2) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 3) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 4) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 5) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 6) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 7) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		}
	      }
	      else {				/* odd column */
		if (bytes == 1) {
		  val = val & 0x1ff;		/* drop surrounding bits at right border */
		}
		if (memram[addr] & 0x80) {
		  val8 = colourtabo2[(val >> 1) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 2) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 3) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 4) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 5) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe2[(val >> 6) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo2[(val >> 7) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		}
		else {
		  val8 = colourtabo[(val >> 1) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 2) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 3) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 4) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 5) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabe[(val >> 6) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		  val8 = colourtabo[(val >> 7) & 7];
		  *virtptr++ = val8;
		  *virtptr++ = val8;
		}
	      }
	      addr++;
	    } /* for bytes */
	    return -1;

//	  case MODEHRESRGB :		/* hires RGB */
/*	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehres;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    bit = 0;
	    for (bytes = 20; bytes; bytes--) {
	      val8 = memram[addr-1];
	      if (val8 & 0x80) {
		val = (doublebits[val8 & 0x7f] << 1) | (bit == 0x40);
	      }
	      else {
		val = doublebits[val8];
	      }
	      bit = val8 & 0x40;

	      val8 = memram[addr++];
	      if (val8 & 0x80) {
		val = (doublebits[val8 & 0x7f] << 1) | (bit == 0x40);
	      }
	      else {
		val = doublebits[val8];
	      }
	      bit = val8 & 0x40;
	      val8 = memram[addr++];
	      if (val8 & 0x80) {
		val = val | (doublebits[val8 & 0x7f] << 15) | (bit << 8);
	      }
	      else {
		val = val | (doublebits[val8] << 14);
	      }
	      bit = val8 & 0x40;
	      val8 = memram[addr];
	      if (val8 & 0x80) {
		val = val | (doublebits[val8 & 0x7f] << 29) | (bit << 22);
	      }
	      else {
		val = val | doublebits[val8] << 28;
	      }
	      if (bytes == 20) { val = val & 0x07ffffff8; }
//	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = dhrscol4rgb[ val & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >>  1) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >>  2) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >>  3) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >>  4) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >>  5) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >>  6) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >>  7) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >>  8) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >>  9) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 10) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 11) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 12) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 13) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 14) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 15) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 16) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 17) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 18) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 19) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 20) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 21) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 22) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 23) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 24) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 25) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 26) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 27) & 0xf ];
	    } // for bytes //
	    return -1;
*/
	  case MODEHRESBLUR :		/* hires blur */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehres;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    bit = 0;
	    for (bytes = 20; bytes; bytes--) {
	      val8 = memram[addr++];
	      if (val8 & 0x80) {
		val = (doublebits[val8 & 0x7f] << 1) | (bit == 0x40);
	      }
	      else {
		val = doublebits[val8];
	      }
	      bit = val8 & 0x40;
	      val8 = memram[addr++];
	      if (val8 & 0x80) {
		val = val | (doublebits[val8 & 0x7f] << 15) | (bit << 8);
	      }
	      else {
		val = val | (doublebits[val8] << 14);
	      }
	      bit = val8 & 0x40;
	      val8 = memram[addr];
	      if (val8 & 0x80) {
		val = val | (doublebits[val8 & 0x7f] << 29) | (bit << 22);
	      }
	      else {
		val = val | doublebits[val8] << 28;
	      }
	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = hrescol4[ val & 0xf ];
	      *virtptr++ = hrescol [ (val >>  1) & 0xf ];
	      *virtptr++ = hrescol2[ (val >>  2) & 0xf ];
	      *virtptr++ = hrescol3[ (val >>  3) & 0xf ];
	      *virtptr++ = hrescol4[ (val >>  4) & 0xf ];
	      *virtptr++ = hrescol [ (val >>  5) & 0xf ];
	      *virtptr++ = hrescol2[ (val >>  6) & 0xf ];
	      *virtptr++ = hrescol3[ (val >>  7) & 0xf ];
	      *virtptr++ = hrescol4[ (val >>  8) & 0xf ];
	      *virtptr++ = hrescol [ (val >>  9) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 10) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 11) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 12) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 13) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 14) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 15) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 16) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 17) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 18) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 19) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 20) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 21) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 22) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 23) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 24) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 25) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 26) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 27) & 0xf ];
	    } /* for bytes */
	    return -1;

	  case MODEHRESOFF :		/* illegal hires standard */
	  case MODEHRESRGBOFF :		/* illegal hires RGB */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehreso;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val = ((memram[addr-1] & 0x60) >> 5)
		  | ((memram[addr]   & 0x7f) << 2)
		  | ((memram[addr+1] & 0x3)  << 9);
	      if ((bytes & 1) == 0) {              /* even column */
		if (bytes == 40) {
		  val = val & 0x7fc;       /* drop surrounding bits at left border */
		}
		val8 = colourtabe[(val >> 1) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 2) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 3) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 4) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 5) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 6) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 7) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
	      }
	      else {				/* odd column */
		if (bytes == 1) {
		  val = val & 0x1ff;		/* drop surrounding bits at right border */
		}
		val8 = colourtabo[(val >> 1) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 2) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 3) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 4) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 5) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabe[(val >> 6) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
		val8 = colourtabo[(val >> 7) & 7];
		*virtptr++ = val8;
		*virtptr++ = val8;
	      }
	      addr++;
	    } /* for bytes */
	    return -1;

	  case MODEHRESBLUROFF :	/* illegal hires blur */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr];
	    mode = virtcachemode | cachehreso;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val8 = memram[addr++] & 0x7f;
	      val = doublebits[val8];
	      val8 = memram[addr++] & 0x7f;
	      val = val | (doublebits[val8] << 14);
	      val8 = memram[addr] & 0x7f;
	      val = val | doublebits[val8] << 28;
	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = hrescol4[ val & 0xf ];
	      *virtptr++ = hrescol [ (val >>  1) & 0xf ];
	      *virtptr++ = hrescol2[ (val >>  2) & 0xf ];
	      *virtptr++ = hrescol3[ (val >>  3) & 0xf ];
	      *virtptr++ = hrescol4[ (val >>  4) & 0xf ];
	      *virtptr++ = hrescol [ (val >>  5) & 0xf ];
	      *virtptr++ = hrescol2[ (val >>  6) & 0xf ];
	      *virtptr++ = hrescol3[ (val >>  7) & 0xf ];
	      *virtptr++ = hrescol4[ (val >>  8) & 0xf ];
	      *virtptr++ = hrescol [ (val >>  9) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 10) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 11) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 12) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 13) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 14) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 15) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 16) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 17) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 18) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 19) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 20) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 21) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 22) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 23) & 0xf ];
	      *virtptr++ = hrescol4[ (val >> 24) & 0xf ];
	      *virtptr++ = hrescol [ (val >> 25) & 0xf ];
	      *virtptr++ = hrescol2[ (val >> 26) & 0xf ];
	      *virtptr++ = hrescol3[ (val >> 27) & 0xf ];
	    } /* for bytes */
	    return -1;

	  case MODEDHRSMONO :		/* double hires mono */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachedhrs;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 40; bytes; bytes--) {
	      val = memram[addr+0x10000];
	      for ( bit = 1; bit<128; bit=bit<<1) {
		*virtptr++ = (val & bit) ? virtmonocol1 : virtmonocol0;
	      } /* for bit */
	      val = memram[addr++];
	      for ( bit = 1; bit<128; bit=bit<<1) {
		*virtptr++ = (val & bit) ? virtmonocol1 : virtmonocol0;
	      } /* for bit */
	    } /* for bytes */
	    return -1;

	  case MODEDHRS :		/* double hires */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachedhrs;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val = ( memram[addr+0x10000]&0x7f)
		  | ((memram[addr]        &0x7f) << 7)
		  | ((memram[addr+0x10001]&0x7f) << 14)
		  | ((memram[addr+1]      &0x7f) << 21)
		  | ((memram[addr+0x10002]&0x7f) << 28);
	      if (bytes == 1) { val = val & 0x0fffffff; }
	      *virtptr++ = dhrscol [ val & 0xf ];
	      *virtptr++ = dhrscol2[ (val >>  1) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >>  2) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >>  3) & 0xf ];
	      *virtptr++ = dhrscol [ (val >>  4) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >>  5) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >>  6) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >>  7) & 0xf ];
	      *virtptr++ = dhrscol [ (val >>  8) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >>  9) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >> 10) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >> 11) & 0xf ];
	      *virtptr++ = dhrscol [ (val >> 12) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >> 13) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >> 14) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >> 15) & 0xf ];
	      *virtptr++ = dhrscol [ (val >> 16) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >> 17) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >> 18) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >> 19) & 0xf ];
	      *virtptr++ = dhrscol [ (val >> 20) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >> 21) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >> 22) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >> 23) & 0xf ];
	      *virtptr++ = dhrscol [ (val >> 24) & 0xf ];
	      *virtptr++ = dhrscol2[ (val >> 25) & 0xf ];
	      *virtptr++ = dhrscol3[ (val >> 26) & 0xf ];
	      *virtptr++ = dhrscol4[ (val >> 27) & 0xf ];
	      addr++;
	      addr++;
	    } /* for (bytes) */
	    return -1;

	  case MODEDHRS140 :		/* double hires 140*192, just a joke */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachedhrs;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val8 = dhrscol[ (memram[addr+0x10000]&0xf)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      val8 = dhrscol[ ((memram[addr+0x10000]&0x70) >> 4) | ((memram[addr] & 0x01) << 3)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      val8 = dhrscol[ ((memram[addr]&0x1e) >> 1)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      val8 = dhrscol[ ((memram[addr]&0x60) >> 5) | ((memram[addr+0x10000+1] & 0x03) << 2)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      addr++;
	      val8 = dhrscol[ ((memram[addr+0x10000]&0x3c) >> 2)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      val8 = dhrscol[ ((memram[addr+0x10000]&0x40) >> 6) | ((memram[addr] & 0x7) << 1)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      val8 = dhrscol[ ((memram[addr]&0x78) >> 3)];
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      *virtptr++ = val8;
	      addr++;
	    } /* for (bytes) */
	    return -1;

	  case MODEDHRSRGB :		/* double hires RGB */
	    addr = hresaddr[rastline] + virthrespage;
	    virtvideobyte = memram[addr+65536];
	    mode = virtcachemode | cachedhrs;
	    if (virtcache[rastline] == mode) { return 0; }
	    virtcache[rastline] = mode;
	    virtcopy = 1;
	    for (bytes = 20; bytes; bytes--) {
	      val = ( memram[addr+0x10000]&0x7f)
		  | ((memram[addr]        &0x7f) << 7)
		  | ((memram[addr+0x10001]&0x7f) << 14)
		  | ((memram[addr+1]      &0x7f) << 21)
		  | ((memram[addr+0x10002]&0x7f) << 28);
	      val = (val << 3) | ((memram[addr-1] & 0x70) >> 4);
	      if (bytes == 20) { val = val & 0x07ffffff8; }
	      *virtptr++ = dhrscol1rgb[ val & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >>  1) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >>  2) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >>  3) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >>  4) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >>  5) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >>  6) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >>  7) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >>  8) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >>  9) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 10) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 11) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 12) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 13) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 14) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 15) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 16) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 17) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 18) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 19) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 20) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 21) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 22) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 23) & 0xf ];
	      *virtptr++ = dhrscol1rgb[ (val >> 24) & 0xf ];
	      *virtptr++ = dhrscol2rgb[ (val >> 25) & 0xf ];
	      *virtptr++ = dhrscol3rgb[ (val >> 26) & 0xf ];
	      *virtptr++ = dhrscol4rgb[ (val >> 27) & 0xf ];
	      addr++;
	      addr++;
	    } /* for (bytes) */
	    return -1;

	} /* switch */

	return 0;	/* not needed, only here to avoid compiler warnings */

      } /* virtcalcline */

	    
/*-------------------------------------*/


      void virtline(unsigned int rastline) {
	unsigned char *sourceptr;
	register unsigned char *destptr;
	register unsigned int i;

	if (virtdouble) {
	  virtcache[rastline] = 0;
	  virttextpage = 0x400;
	  virthrespage = 0x2000;
	  virtcalcline(imagegetaddress(window, rastline + 8)       + 40, rastline);
	  virtcache[rastline] = 0;
	  virttextpage = 0x800;
	  virthrespage = 0x4000;
	  virtcalcline(imagegetaddress(window, rastline + 192 + 8) + 40, rastline);
	}
	else {
	  if (virtscanline == 0) {
	    sourceptr = imagegetaddress(window, (rastline <<1) + 8) + 40;
	    if (virtcalcline(sourceptr, rastline)) {
	      memcpy( imagegetaddress(window, (rastline <<1) + 9) + 40, sourceptr, 560);
	    }
	  }
	  else {
	    if (virtscanline == 1) {
	      sourceptr = imagegetaddress(window, (rastline <<1) + 8) + 40;
	      if (virtcalcline(sourceptr, rastline)) {
	        destptr   = imagegetaddress(window, (rastline <<1) + 9) + 40;
	        for (i=560; i; i--) {
	          *(destptr++) = (*sourceptr++) + 0x20;
	        } /* for (i) */
	      }
	    }
	    else {
	      if (virtcalcline(imagegetaddress(window, (rastline <<1) + 8) + 40, rastline)) {
	        memset(imagegetaddress(window, (rastline <<1) + 9) + 40, 0, 560);
	      }
	    }
	  }
	}

      } /* virtline */

	    
/*-------------------------------------*/


      void virtscreencopy () {

	if (virtcopy) {
	  imageupdate();
	  virtcopy = 0;
	}
	else {
	  imagenoupdate();
	}

      } /* virtscreencopy */


/*-------------------------------------*/


      void virtwrite0400 (unsigned int addr) {
	register unsigned int line;
	register unsigned int *pointer;

	if (~(virtcachemode & cachepage)) {
	  line = textline[addr - 0x400];
	  if ((line >= virtsplit) && (line <= 191)) {
	    pointer = &virtcache[line];
	    *pointer = 0;
	    *(pointer+1) = 0;
	    *(pointer+2) = 0;
	    *(pointer+3) = 0;
	    *(pointer+4) = 0;
	    *(pointer+5) = 0;
	    *(pointer+6) = 0;
	    *(pointer+7) = 0;
/*	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line] = 0;
*/	  }
	}
      } /* virtwrite0400 */


/*-------------------------------------*/


      void virtwrite0400aux (unsigned int addr) {
	register unsigned int line;

	if (virt80col) {
	  if (~(virtcachemode & cachepage)) {
	    line = textline[addr - 0x400];
	    if ((line >= virtsplit) && (line <= 191)) {
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line] = 0;
	    }
	  }
	}
      } /* virtwrite0400aux */


/*-------------------------------------*/


      void virtwrite0800 (unsigned int addr) {
	register unsigned int line;

	if (virtcachemode & cachepage) {
	  line = textline[addr - 0x800];
	  if ((line >= virtsplit) && (line <= 191)) {
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line++] = 0;
	    virtcache[line] = 0;
	  }
	}
      } /* virtwrite0800 */


/*-------------------------------------*/


      void virtwrite0800aux (unsigned int addr) {
	register unsigned int line;

	if (virt80col) {
	  if (virtcachemode & cachepage) {
	    line = textline[addr - 0x800];
	    if ((line >= virtsplit) && (line <= 191)) {
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line++] = 0;
	      virtcache[line] = 0;
	    }
	  }
	}
      } /* virtwrite0800aux */


/*-------------------------------------*/


      void virtwrite2000 (unsigned int addr) {
	register unsigned int line;

	if (!(virtcachemode & cachepage)) {
	  line = hresline[addr - 0x2000];
	  if (line < virtsplit) {
	    virtcache[line] = 0;
	  }
	}
      } /* virtwrite2000 */


/*-------------------------------------*/


      void virtwrite2000aux (unsigned int addr) {
	register unsigned int line;

	if (virtdhres) {
	  if (!(virtcachemode & cachepage)) {
	    line = hresline[addr - 0x2000];
	    if (line < virtsplit) {
	      virtcache[line] = 0;
	    }
	  }
	}
      } /* virtwrite2000aux */


/*-------------------------------------*/


      void virtwrite4000 (unsigned int addr) {
	register unsigned int line;

	if (virtcachemode & cachepage) {
	  line = hresline[addr - 0x4000];
	  if (line < virtsplit) {
	    virtcache[line] = 0;
	  }
	}
      } /* virtwrite4000 */


/*-------------------------------------*/


      void virtwrite4000aux (unsigned int addr) {
	register unsigned int line;

	if (virtdhres) {
	  if (virtcachemode & cachepage) {
	    line = hresline[addr - 0x4000];
	    if (line < virtsplit) {
	      virtcache[line] = 0;
	    }
	   }
	}
      } /* virtwrite4000aux */


/*-------------------------------------*/


      void virtmenu () {
	unsigned int screen;
	unsigned int keyboard;
	unsigned int window;
	unsigned char key;
	unsigned char update;

	screen = screenstore();
	if (!windowaddio( -1, -1, WINDOWXSIZE, WINDOWYSIZE, -1, 1, "Video Options", &keyboard, &window)) {

	  messageflag = 0;
	  update = 1;
	  do {
	    if (update) {
	      channelout(window, 12);		/* clear window */
	      stringwrite(window, "\r[ESC] - Quit");

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Monitor:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[C] - Color screen                  ");
	      channelout(window, virtgetmonochrome()==0 ? 42:32);
	      stringwrite(window, "\r[G] - Green screen                  ");
	      channelout(window, virtgetmonochrome()==2 ? 42:32);
	      stringwrite(window, "\r[M] - Monochrome screen (white)     ");
	      channelout(window, virtgetmonochrome()==1 ? 42:32);
	      stringwrite(window, "\r[A] - Amber screen                  ");
	      channelout(window, virtgetmonochrome()==3 ? 42:32);

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Text mode:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[T] - ");
	      if (virtgettextmode() == 0) {
	        stringwrite(window, "Standard text display");
	      }
	      else {
	        stringwrite(window, "Blur text display");
	      }

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Hi-res mode:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[H] - ");
	      if (virtgethresmode() == 0) {
	        stringwrite(window, "Standard hires display");
	      }
	      else {
		if (virtgethresmode() == 1) {
		  stringwrite(window, "RGB hires display");
		}
		else {
		  stringwrite(window, "Blur hires display");
		}
	      }

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Double Hi-res mode:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[D] - ");
	      if (virtgetdhrsmode() == 0) {
	        stringwrite(window, "560 * 192 resolution");
	      }
	      else {
	        if (virtgetdhrsmode() == 1) {
		  stringwrite(window, "140 * 192 resolution");
		}
		else {
		  stringwrite(window, "RGB");
		}
	      }

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Lo-res/Hi-res color palette:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[0] - Palette 0 (Apple IIgs)        ");
	      channelout(window, virtgetpalette()==0 ? 42:32);
	      stringwrite(window, "\r[1] - Palette 1 (Standard)          ");
	      channelout(window, virtgetpalette()==1 ? 42:32);
	      stringwrite(window, "\r[2] - Palette 2 (ApplePC-like)      ");
	      channelout(window, virtgetpalette()==2 ? 42:32);
              stringwrite(window, "\r[3] - Palette 3 (Grayscale)         ");
	      channelout(window, virtgetpalette()==3 ? 42:32);

	      stringwrite(window, "\r\r      ");
	      imagesettextattribute(window, IMAGEUNDERLINE);
	      stringwrite(window, "Display:");
	      imagesettextattribute(window, 0);
	      imagesetcursor(window, imagegetcursorx(window), imagegetcursory(window)+2);
	      stringwrite(window, "\r[V] - View ");
	      if (virtdouble) {
	        stringwrite(window, "both pages simultaneously");
	      }
	      else {
	        stringwrite(window, "one page");
	      }
	      stringwrite(window, "\r[L] - Scanline: ");
	      if (virtscanline == 0) {
	        stringwrite(window, "None");
	      }
	      else {
	        if (virtscanline == 1) {
	          stringwrite(window, "HalfTone");
	        }
	        else {
	          stringwrite(window, "Full");
	        }
	      }

	      stringwrite(window,"\r[F] - Font: ");
	      switch (virtgetcharset()) {
	        case USA :
	          stringwrite(window, "USA");
	          break;
	        case France :
	          stringwrite(window, "France");
	          break;
	        case Germany :
	          stringwrite(window, "Germany");
	          break;
	        case UK :
	          stringwrite(window, "United Kingdom");
	          break;
	        case Denmark1 :
	          stringwrite(window, "Denmark1");
	          break;
	        case Sweden :
	          stringwrite(window, "Sweden");
	          break;
	        case Italy :
	          stringwrite(window, "Italy");
	          break;
	        case Spain :
	          stringwrite(window, "Spain");
	          break;
	        case Japan :
	          stringwrite(window, "Japan");
	          break;
	        case Norway :
	          stringwrite(window, "Norway");
	          break;
	        case Denmark2 :
	          stringwrite(window, "Denmark2");
	          break;
	        default :
	          stringwrite(window, "None selected");
	          break;
	      }

	      update = 0;
	    }
	    do {
	      imageupdate();
	      key = (unsigned char)channelin(keyboard);
	    }
	    while ((key == 0) && (!exitprogram));
	    switch (key) {
	      case 'c' :
	      case 'C' :
		virtsetmonochrome(0);
		update = 1;
		break;
	      case 'g' :
	      case 'G' :
		virtsetmonochrome(2);
		update = 1;
		break;
	      case 'm' :
	      case 'M' :
		virtsetmonochrome(1);
		update = 1;
	        break;
	      case 'a' :
	      case 'A' :
		virtsetmonochrome(3);
		update = 1;
		break;

	      case 't' :
	      case 'T' :
	        update = virtgettextmode();
	        update++;
	        if (update > 1) { update = 0; }
	        virtsettextmode(update);
	        update = 1;
	        break;

	      case 'h' :
	      case 'H' :
	        update = virtgethresmode();
	        update++;
	        if (update > 2) { update = 0; }
	        virtsethresmode(update);
	        update = 1;
	        break;

	      case 'd' :
	      case 'D' :
	        update = virtgetdhrsmode();
	        update++;
	        if (update > 2) { update = 0; }
	        virtsetdhrsmode(update);
	        update = 1;
	        break;

	      case '0' :
		virtsetpalette(0);
		update = 1;
		break;
	      case '1' :
		virtsetpalette(1);
		update = 1;
		break;
	      case '2' :
		virtsetpalette(2);
		update = 1;
		break;
	      case '3' :
		virtsetpalette(3);
		update = 1;
		break;

	      case 'v' :
	      case 'V' :
	        virtdouble = !virtdouble;
	        virtcacheinit();
	        virtsetmode();
		update = 1;
		break;
	      case 'f' :
	      case 'F' :
	        update = virtgetcharset();
	        update++;
	        if (update > Denmark2) { update = USA; }
	        virtsetcharset(update);
	        update = 1;
	        break;
	      case 'l' :
	      case 'L' :
	        virtscanline++;
	        if (virtscanline > 2) { virtscanline = 0; }
	        virtcacheinit();
		update = 1;
		break;
	    } /* switch (key) */
	  }
	  while ((key != 27) && (key != 32) && (key != 13) && (!exitprogram));
	  channelclose(keyboard);
	  channelclose(window);
	  messageflag = 1;
	}
	screenrestore(screen);

      } /* virtmenu */
